"""Classes for computing pitch features"""
from __future__ import annotations

import pathlib
import typing

import librosa
import numpy as np

from _kalpy import feat
from _kalpy.matrix import CompressedMatrix, FloatMatrixBase
from _kalpy.util import BaseFloatMatrixWriter, CompressedMatrixWriter
from kalpy.data import Segment
from kalpy.utils import generate_write_specifier


class PitchComputer:
    """
    Class for computing pitch features

    Parameters
    ----------
    sample_frequency: float
        Sample rate to use in generating features, audio will be resampled as necessary,
        defaults to 16000
    frame_length: int
        Frame length in milliseconds, defaults to 25
    frame_shift: int
        Frame shift in milliseconds, defaults to 10
    min_f0: float
        Minimum f0 to search (Hz)
    max_f0: float
        Maximum f0 to search (Hz)
    soft_min_f0: float
        Minimum f0, applied in a soft way, must not exceed min_f0
    penalty_factor: float
        Cost factor for f0 change
    lowpass_cutoff: float
        Cutoff frequency for lowpass filter
    resample_frequency: float
        Frequency that we down-sample the signal to.  Must be more than twice lowpass-cutoff
    delta_pitch: float
        Smallest relative change in pitch that our algorithm measures
    nccf_ballast: float
        Increasing this factor reduces NCCF for quiet frames, helping ensure pitch
        continuity in unvoiced region
    lowpass_filter_width: int
        Integer that determines filter width of lowpass filter
    upsample_filter_width: int
        Integer that determines filter width when upsampling NCCF
    max_frames_latency: int
        The maximum number of frames of latency that we allow the pitch-processing
        to introduce, for online operation. If you set this to a large value,
        there would be no inaccuracy from the Viterbi traceback (but it might make
        you wait to see the pitch). This is not very relevant for the online
        operation: normalization-right-context is more relevant, you
        can just leave this value at zero.
    frames_per_chunk: int
        Only relevant for the function ComputeKaldiPitch which is called by
        compute-kaldi-pitch-feats. If nonzero, we provide the input as chunks of
        this size. This affects the energy normalization which has a small effect
        on the resulting features, especially at the beginning of a file. For best
        compatibility with online operation (e.g. if you plan to train models for
        the online-decoding setup), you might want to set this to a small value,
        like one frame.
    simulate_first_pass_online: bool
        If true, compute-kaldi-pitch-feats will output features that correspond to
        what an online decoder would see in the first pass of decoding-- not the final
        version of the features, which is the default.  Relevant if --frames-per-chunk > 0
    recompute_frame: int
        Only relevant for online operation or when emulating online operation
        (e.g. when setting frames_per_chunk). This is the frame-index on which we
        recompute the NCCF (e.g. frame-index 500 = after 5 seconds); if the
        segment ends before this we do it when the segment ends. We do this by
        re-computing the signal average energy, which affects the NCCF via the
        "ballast term", scaling the resampled NCCF by a factor derived from the
        average change in the "ballast term", and re-doing the backtrace
        computation. Making this infinity would be the most exact, but would
        introduce unwanted latency at the end of long utterances, for little
        benefit.
    snip_edges: bool
        Flag for whether edges of segments should be cutoff to ensure no zero padding in frames,
        defaults to True
    pitch_scale: float
        Scaling factor for the final normalized log-pitch valu
    pov_scale: float
        Scaling factor for final POV (probability of voicing) feature
    pov_offset: float
        This can be used to add an offset to the POV feature. Intended for use in
        online decoding as a substitute for  CMN.
    delta_pitch_scale: float
        Term to scale the final delta log-pitch feature
    delta_pitch_noise_stddev: float
        Standard deviation for noise we add to the delta log-pitch (before scaling);
        should be about the same as delta-pitch option to pitch creation.
        The purpose is to get rid of peaks in the delta-pitch caused by discretization of
        pitch values.
    normalization_left_context: int
        Left-context (in frames) for moving window normalization
    normalization_right_context: int
        Right-context (in frames) for moving window normalization
    delta_window: int
        Number of frames on each side of central frame, to use for delta window.
    delay: int
        Number of frames by which the pitch information is delayed.
    add_pov_feature: bool
        If true, the warped NCCF is added to output features
    add_normalized_log_pitch: bool
        If true, the log-pitch with POV-weighted mean subtraction over 1.5 second window
        is added to output features
    add_delta_pitch: bool
        If true, time derivative of log-pitch is added to output features
    add_raw_log_pitch: bool
        If true, log(pitch) is added to output features
    """

    def __init__(
        self,
        sample_frequency: float = 16000,
        frame_length: int = 25,
        frame_shift: int = 10,
        min_f0: float = 50,
        max_f0: float = 400,
        soft_min_f0: float = 10.0,
        penalty_factor: float = 0.1,
        lowpass_cutoff: float = 1000,
        resample_frequency: float = 4000,
        delta_pitch: float = 0.005,
        nccf_ballast: float = 7000,
        lowpass_filter_width: int = 1,
        upsample_filter_width: int = 5,
        max_frames_latency: int = 0,
        frames_per_chunk: int = 0,
        simulate_first_pass_online: bool = False,
        recompute_frame: int = 500,
        snip_edges: bool = False,
        pitch_scale: float = 2.0,
        pov_scale: float = 2.0,
        pov_offset: float = 0.0,
        delta_pitch_scale: float = 10.0,
        delta_pitch_noise_stddev: float = 0.005,
        normalization_left_context: int = 75,
        normalization_right_context: int = 75,
        delta_window: int = 2,
        delay: int = 0,
        add_pov_feature: bool = True,
        add_normalized_log_pitch: bool = True,
        add_delta_pitch: bool = True,
        add_raw_log_pitch: bool = False,
    ):
        self.extraction_opts = feat.PitchExtractionOptions()
        self.extraction_opts.samp_freq = sample_frequency
        self.extraction_opts.frame_length_ms = frame_length
        self.extraction_opts.frame_shift_ms = frame_shift
        self.extraction_opts.min_f0 = min_f0
        self.extraction_opts.max_f0 = max_f0
        self.extraction_opts.soft_min_f0 = soft_min_f0
        self.extraction_opts.penalty_factor = penalty_factor
        self.extraction_opts.lowpass_cutoff = lowpass_cutoff
        self.extraction_opts.resample_freq = resample_frequency
        self.extraction_opts.delta_pitch = delta_pitch
        self.extraction_opts.nccf_ballast = nccf_ballast
        self.extraction_opts.lowpass_filter_width = lowpass_filter_width
        self.extraction_opts.upsample_filter_width = upsample_filter_width
        self.extraction_opts.max_frames_latency = max_frames_latency
        self.extraction_opts.frames_per_chunk = frames_per_chunk
        self.extraction_opts.simulate_first_pass_online = simulate_first_pass_online
        self.extraction_opts.recompute_frame = recompute_frame
        self.extraction_opts.snip_edges = snip_edges

        self.process_opts = feat.ProcessPitchOptions()
        self.process_opts.pitch_scale = pitch_scale
        self.process_opts.pov_scale = pov_scale
        self.process_opts.pov_offset = pov_offset
        self.process_opts.delta_pitch_scale = delta_pitch_scale
        self.process_opts.delta_pitch_noise_stddev = delta_pitch_noise_stddev
        self.process_opts.normalization_left_context = normalization_left_context
        self.process_opts.normalization_right_context = normalization_right_context
        self.process_opts.delta_window = delta_window
        self.process_opts.delay = delay
        self.process_opts.add_pov_feature = add_pov_feature
        self.process_opts.add_normalized_log_pitch = add_normalized_log_pitch
        self.process_opts.add_delta_pitch = add_delta_pitch
        self.process_opts.add_raw_log_pitch = add_raw_log_pitch

    @property
    def parameters(self):
        return {
            "sample_frequency": self.extraction_opts.samp_freq,
            "frame_length": self.extraction_opts.frame_length_ms,
            "frame_shift": self.extraction_opts.frame_shift_ms,
            "min_f0": self.extraction_opts.min_f0,
            "max_f0": self.extraction_opts.max_f0,
            "soft_min_f0": self.extraction_opts.soft_min_f0,
            "penalty_factor": self.extraction_opts.penalty_factor,
            "lowpass_cutoff": self.extraction_opts.lowpass_cutoff,
            "resample_frequency": self.extraction_opts.resample_freq,
            "delta_pitch": self.extraction_opts.delta_pitch,
            "nccf_ballast": self.extraction_opts.nccf_ballast,
            "lowpass_filter_width": self.extraction_opts.lowpass_filter_width,
            "upsample_filter_width": self.extraction_opts.upsample_filter_width,
            "max_frames_latency": self.extraction_opts.max_frames_latency,
            "frames_per_chunk": self.extraction_opts.frames_per_chunk,
            "simulate_first_pass_online": self.extraction_opts.simulate_first_pass_online,
            "recompute_frame": self.extraction_opts.recompute_frame,
            "snip_edges": self.extraction_opts.snip_edges,
            "pitch_scale": self.process_opts.pitch_scale,
            "pov_scale": self.process_opts.pov_scale,
            "pov_offset": self.process_opts.pov_offset,
            "delta_pitch_scale": self.process_opts.delta_pitch_scale,
            "delta_pitch_noise_stddev": self.process_opts.delta_pitch_noise_stddev,
            "normalization_left_context": self.process_opts.normalization_left_context,
            "normalization_right_context": self.process_opts.normalization_right_context,
            "delta_window": self.process_opts.delta_window,
            "delay": self.process_opts.delay,
            "add_pov_feature": self.process_opts.add_pov_feature,
            "add_normalized_log_pitch": self.process_opts.add_normalized_log_pitch,
            "add_delta_pitch": self.process_opts.add_delta_pitch,
            "add_raw_log_pitch": self.process_opts.add_raw_log_pitch,
        }

    def compute_pitch(
        self,
        segment: Segment,
    ) -> np.ndarray:
        """
        Compute pitch features for a segment

        Parameters
        ----------
        segment: :class:`~kalpy.feat.mfcc.Segment`
            Acoustic segment to generate pitch features

        Returns
        -------
        :class:`numpy.ndarray`
            Feature matrix for the segment
        """
        pitch = self.compute_pitch_for_export(segment, compress=False)
        return pitch.numpy()

    def compute_pitch_for_export(
        self,
        segment: Segment,
        compress: bool = True,
    ) -> FloatMatrixBase:
        """
        Generate pitch features for exporting to a kaldi archive

        Parameters
        ----------
        segment: :class:`~kalpy.feat.mfcc.Segment`
            Acoustic segment to generate pitch features

        Returns
        -------
        :class:`_kalpy.matrix.FloatMatrixBase`
            Feature matrix for the segment
        """
        duration = None
        if segment.end is not None and segment.begin is not None:
            duration = segment.end - segment.begin
        wave, sr = librosa.load(
            segment.file_path,
            sr=16000,
            offset=segment.begin,
            duration=duration,
            mono=False,
        )
        wave = np.round(wave * 32768)
        if len(wave.shape) == 2:
            channel = 0 if segment.channel is None else segment.channel
            wave = wave[channel, :]
        pitch = self.compute_pitch_for_wave(wave)
        if compress:
            pitch = CompressedMatrix(pitch)
        return pitch

    def compute_pitch_for_wave(
        self,
        wave: np.ndarray,
    ) -> FloatMatrixBase:
        """
        Generate pitch features for exporting to a kaldi archive

        Parameters
        ----------
        wave: :class:`~numpy.ndarray`
            Waveform

        Returns
        -------
        :class:`_kalpy.matrix.FloatMatrixBase`
            Feature matrix for the segment
        """
        pitch = feat.compute_pitch(wave, self.extraction_opts, self.process_opts)
        return pitch

    def export_feats(
        self,
        file_name: typing.Union[pathlib.Path, str],
        segments: typing.Iterable[typing.Tuple[str, Segment]],
        write_scp: bool = False,
        compress: bool = True,
    ) -> None:
        """
        Export features to a kaldi archive file (i.e., pitch.ark)

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Archive file path to export to
        segments: dict[str, :class:`kalpy.feat.mfcc.Segment`]
            Mapping of utterance IDs to Segment objects
        write_scp: bool
            Flag for whether an SCP file should be generated as well
        compress: bool
            Flag for whether to export features as a compressed archive
        """
        write_specifier = generate_write_specifier(file_name, write_scp)
        if compress:
            writer = CompressedMatrixWriter(write_specifier)
        else:
            writer = BaseFloatMatrixWriter(write_specifier)
        for key, segment in segments:
            feats = self.compute_pitch_for_export(segment, compress=compress)
            writer.Write(str(key), feats)
        writer.Close()



"""Classes for computing LDA transforms"""
from __future__ import annotations

import logging
import pathlib
import typing

from _kalpy import transform
from _kalpy.util import ConstIntegerSet, Output
from kalpy.feat.data import FeatureArchive
from kalpy.gmm.data import AlignmentArchive
from kalpy.gmm.utils import read_gmm_model

logger = logging.getLogger("kalpy.lda")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class LdaStatsAccumulator:
    def __init__(
        self,
        acoustic_model_path: typing.Union[pathlib.Path, str],
        silence_phones: typing.List[int],
        weight_distribute: bool = False,
        rand_prune: float = 0.0,
    ):
        self.acoustic_model_path = str(acoustic_model_path)
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.silence_phones = silence_phones
        self.rand_prune = rand_prune
        self.weight_distribute = weight_distribute
        self.lda = transform.LdaEstimate()

    def accumulate_stats(
        self,
        feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
        callback: typing.Callable = None,
    ):
        silence_weight = 0.0
        silence_set = ConstIntegerSet(self.silence_phones)
        num_done = 0
        for utterance_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utterance_id} due to zero-length features")
                continue
            try:
                alignment = alignment_archive[utterance_id]
            except KeyError:
                logger.warning(f"Skipping {utterance_id} due to missing alignment")
                continue
            if self.lda.Dim() == 0:
                self.lda.Init(self.transition_model.NumPdfs(), feats.NumCols())
            if callback:
                callback(alignment.utterance_id)
            self.lda.acc_lda(
                self.transition_model,
                alignment.alignment,
                feats,
                silence_set,
                rand_prune=self.rand_prune,
                silence_weight=silence_weight,
            )
            num_done += 1
            if num_done % 100 == 0:
                logger.info(f"Done {num_done} utterances.")
        logger.info(f"Done {num_done} files.")

    def export_transform(
        self, file_name: str, feature_archive: FeatureArchive, alignment_archive: AlignmentArchive
    ):
        file_name = str(file_name)
        self.accumulate_stats(feature_archive, alignment_archive)
        ko = Output(file_name, True)
        self.lda.Write(ko.Steam(), True)
        logger.info("Written statistics.")


class MlltStatsAccumulator:
    def __init__(
        self,
        acoustic_model_path: typing.Union[pathlib.Path, str],
        silence_phones: typing.List[int],
        weight_distribute: bool = False,
        rand_prune: float = 0.0,
    ):
        self.acoustic_model_path = str(acoustic_model_path)
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.silence_phones = silence_phones
        self.rand_prune = rand_prune
        self.weight_distribute = weight_distribute
        self.mllt_accs = transform.MlltAccs(self.acoustic_model.Dim(), rand_prune)

    def accumulate_stats(
        self,
        feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
        callback: typing.Callable = None,
    ):
        silence_weight = 0.0
        silence_set = ConstIntegerSet(self.silence_phones)
        num_done = 0
        tot_like = 0.0
        tot_t = 0.0
        for utterance_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utterance_id} due to zero-length features")
                continue
            try:
                alignment = alignment_archive[utterance_id]
            except KeyError:
                logger.warning(f"Skipping {utterance_id} due to missing alignment")
                continue
            if callback:
                callback(alignment.utterance_id)
            tot_like_this_file, tot_weight_this_file = self.mllt_accs.gmm_acc_mllt(
                self.acoustic_model,
                self.transition_model,
                alignment.alignment,
                feats,
                silence_set,
                silence_weight=silence_weight,
            )
            num_done += 1
            tot_like += tot_like_this_file
            tot_t += tot_weight_this_file
            if tot_weight_this_file != 0:
                logger.info(
                    f"Average like for this file is {tot_like_this_file/tot_weight_this_file} "
                    f"over {tot_weight_this_file} frames."
                )
            if num_done % 10 == 0 and tot_t != 0:
                logger.info(f"Average per frame so far is {tot_like/tot_t}")
        logger.info(f"Done {num_done} files.")
        if tot_t != 0:
            logger.info(
                f"Overall avg like per frame (Gaussian only) = {tot_like/tot_t} over {tot_t} frames."
            )

    def export_stats(
        self, file_name: str, feature_archive: FeatureArchive, alignment_archive: AlignmentArchive
    ):
        file_name = str(file_name)
        self.accumulate_stats(feature_archive, alignment_archive)
        ko = Output(file_name, True)
        self.mllt_accs.Write(ko.Steam(), True)
        logger.info(f"Written accs.")



"""Classes for GMM alignment"""
from __future__ import annotations

import logging
import pathlib
import typing

from _kalpy.fstext import VectorFst
from _kalpy.lat import (
    CompactLattice,
    CompactLatticeWriter,
    ComposeLatticePrunedOptions,
    lm_rescore,
    lm_rescore_carpa,
)
from _kalpy.lm import ConstArpaLm
from kalpy.gmm.data import LatticeArchive
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.lm")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class LmRescorer:
    def __init__(
        self,
        g_fst: VectorFst,
        acoustic_scale: float = 0.1,
        lm_scale: float = 1.0,
        lattice_compose_beam: float = 6.0,
        max_arcs: int = 100000,
        initial_num_arcs: int = 100,
        growth_ratio: float = 1.5,
    ):
        self.g_fst = g_fst
        self.acoustic_scale = acoustic_scale
        self.lm_scale = lm_scale
        self.num_done = 0
        self.num_error = 0
        self.options = ComposeLatticePrunedOptions()
        self.options.lattice_compose_beam = lattice_compose_beam
        self.options.max_arcs = max_arcs
        self.options.initial_num_arcs = initial_num_arcs
        self.options.growth_ratio = growth_ratio

    def rescore_utterance(
        self,
        lattice: CompactLattice,
        add_lm: typing.Union[VectorFst, ConstArpaLm],
        utterance_id: str = None,
    ):
        try:
            if isinstance(add_lm, ConstArpaLm):
                new_lattice = lm_rescore_carpa(
                    lattice,
                    self.g_fst,
                    add_lm,
                    self.options,
                    self.lm_scale,
                    self.acoustic_scale,
                )
            else:
                new_lattice = lm_rescore(
                    lattice,
                    self.g_fst,
                    add_lm,
                    self.options,
                    self.lm_scale,
                    self.acoustic_scale,
                )
            self.num_done += 1
        except Exception:
            self.num_error += 1
            logger.warning(f"Error in rescoring {utterance_id}")
            raise
        return new_lattice

    def rescore_utterances(
        self, lattice_archive: LatticeArchive, add_lm: typing.Union[VectorFst, ConstArpaLm]
    ) -> typing.Generator[CompactLattice]:
        for (utterance_id, lattice) in lattice_archive:
            yield utterance_id, self.rescore_utterance(lattice, add_lm, utterance_id)

    def export_lattices(
        self,
        file_name: typing.Union[str, pathlib.Path],
        lattice_archive: LatticeArchive,
        add_lm: typing.Union[VectorFst, ConstArpaLm],
        write_scp: bool = False,
        callback: typing.Callable = None,
    ):
        write_specifier = generate_write_specifier(file_name, write_scp)
        writer = CompactLatticeWriter(write_specifier)
        for utterance_id, lattice in self.rescore_utterances(lattice_archive, add_lm):
            if callback:
                callback(utterance_id)
            writer.Write(str(utterance_id), lattice)
        writer.Close()



import pickle

import numpy as np
import pytest

from kalpy.data import KaldiMapping
from kalpy.decoder.data import FstArchive
from kalpy.feat.data import FeatureArchive
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.gmm.align import GmmAligner
from kalpy.gmm.data import AlignmentArchive


@pytest.mark.order(3)
def test_align(mono_tree_path, mono_model_path, dictionary_path, mono_temp_dir):
    lc = LexiconCompiler(position_dependent_phones=False)
    lc.load_pronunciations(dictionary_path)

    cmvn_file_name = mono_temp_dir.joinpath("cmvn.ark")
    training_graph_archive = FstArchive(mono_temp_dir.joinpath("fsts.ark"))
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    feature_archive = FeatureArchive(
        mono_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        deltas=True,
    )
    aligner = GmmAligner(mono_model_path, beam=1000, retry_beam=4000)
    for alignment in aligner.align_utterances(training_graph_archive, feature_archive):
        assert alignment.utterance_id == "1-1"
        assert len(alignment.alignment) == 2672
        assert alignment.per_frame_likelihoods.numpy().shape[0] == 2672
        ctm = alignment.generate_ctm(aligner.transition_model, lc.phone_table)
        assert len(ctm) == 243
        ctm_data = pickle.dumps(ctm)
        ctm_new = pickle.loads(ctm_data)
        assert len(ctm_new) == 243


@pytest.mark.order(3)
def test_align_sat_first_pass(
    sat_tree_path,
    sat_align_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    reference_dir,
    align_options,
    reference_first_pass_ali_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones)
    lc.load_pronunciations(sat_dictionary_path)

    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    alignments_file_name = sat_temp_dir.joinpath("ali.ark")
    word_file_name = sat_temp_dir.joinpath("words.ark")
    training_graph_archive = FstArchive(sat_temp_dir.joinpath("fsts.ark"))
    utt2spk = KaldiMapping()
    textgrid_name = sat_temp_dir.joinpath("first_pass.TextGrid")
    utt2spk["1-1"] = "1"
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        splices=True,
    )

    aligner = GmmAligner(sat_align_model_path, **align_options)
    aligner.boost_silence(1.0, lc.silence_symbols)
    aligner.export_alignments(
        alignments_file_name,
        training_graph_archive,
        feature_archive,
        word_file_name=word_file_name,
    )
    assert alignments_file_name.exists()
    alignment_archive = AlignmentArchive(alignments_file_name, words_file_name=word_file_name)
    alignment = alignment_archive["1-1"]
    assert len(alignment.alignment) == 2672
    intervals = alignment.generate_ctm(aligner.transition_model, lc.phone_table)
    ctm = lc.phones_to_pronunciations(alignment.words, intervals)
    ctm.export_textgrid(textgrid_name, file_duration=26.72)
    reference_alignment_archive = AlignmentArchive(reference_first_pass_ali_path)
    reference_alignment = reference_alignment_archive["1-1"]
    phone_ctm = alignment.generate_ctm(aligner.transition_model, lc.phone_table)
    ref_phone_ctm = reference_alignment.generate_ctm(aligner.transition_model, lc.phone_table)
    assert alignment.alignment == reference_alignment.alignment
    assert phone_ctm == ref_phone_ctm


@pytest.mark.order(5)
def test_align_sat_second_pass(
    sat_tree_path,
    sat_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    reference_second_pass_ali_path,
    align_options,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones)
    lc.load_pronunciations(sat_dictionary_path)
    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    trans_file_name = sat_temp_dir.joinpath("trans.ark")
    word_file_name = sat_temp_dir.joinpath("words.ark")
    textgrid_name = sat_temp_dir.joinpath("second_pass.TextGrid")
    alignments_file_name = sat_temp_dir.joinpath("ali_second_pass.ark")
    training_graph_archive = FstArchive(sat_temp_dir.joinpath("fsts.ark"))
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        transform_file_name=trans_file_name,
        splices=True,
    )
    aligner = GmmAligner(sat_model_path, **align_options)
    aligner.boost_silence(1.0, lc.silence_symbols)
    aligner.export_alignments(
        alignments_file_name,
        training_graph_archive,
        feature_archive,
        word_file_name=word_file_name,
    )
    assert alignments_file_name.exists()
    alignment_archive = AlignmentArchive(alignments_file_name, words_file_name=word_file_name)
    alignment = alignment_archive["1-1"]
    assert len(alignment.alignment) == 2672
    intervals = alignment.generate_ctm(aligner.transition_model, lc.phone_table)
    ctm = lc.phones_to_pronunciations(alignment.words, intervals)
    ctm.export_textgrid(textgrid_name, file_duration=26.72)

    reference_alignment_archive = AlignmentArchive(reference_second_pass_ali_path)
    reference_alignment = reference_alignment_archive["1-1"]
    phone_ctm = alignment.generate_ctm(aligner.transition_model, lc.phone_table)
    ref_phone_ctm = reference_alignment.generate_ctm(aligner.transition_model, lc.phone_table)
    assert alignment.alignment == reference_alignment.alignment
    assert phone_ctm == ref_phone_ctm



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Montreal Forced Aligner documentation build configuration file, created by
# sphinx-quickstart on Wed Jun 15 13:27:38 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import date

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.4'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "myst_parser",
    "sphinx.ext.napoleon",
    "sphinx_design",
    "sphinx.ext.viewcode",
]
myst_enable_extensions = ["colon_fence"]
locale_dirs = ["locale/"]  # path is example but recommended.
gettext_compact = False  # optional.
panels_add_bootstrap_css = False
intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
}

xref_links = {
    "mfa_models": ("MFA Models", "https://mfa-models.readthedocs.io/"),
    "anchor": ("Anchor Annotator", "https://anchor-annotator.readthedocs.io/en/latest/"),
    "speechbrain": ("SpeechBrain", "https://speechbrain.github.io/"),
    "scikit-learn": ("scikit-learn", "https://scikit-learn.org/stable/index.html"),
    "click": ("click", "https://click.palletsprojects.com/en/8.1.x/"),
    "pgvector": ("pgvector", "https://github.com/pgvector/pgvector"),
    "pretrained_acoustic_models": (
        "MFA acoustic models",
        "https://mfa-models.readthedocs.io/en/latest/acoustic/index.html",
    ),
    "pretrained_tokenizer_models": (
        "MFA tokenizer models",
        "https://mfa-models.readthedocs.io/en/latest/tokenizer/index.html",
    ),
    "pretrained_dictionaries": (
        "MFA dictionaries",
        "https://mfa-models.readthedocs.io/en/latest/dictionary/index.html",
    ),
    "pretrained_g2p": (
        "MFA G2P models",
        "https://mfa-models.readthedocs.io/en/latest/g2p/index.html",
    ),
    "pretrained_ivector_extractor": (
        "MFA G2P models",
        "https://mfa-models.readthedocs.io/en/latest/ivector/index.html",
    ),
    "pretrained_language_models": (
        "MFA language models",
        "https://mfa-models.readthedocs.io/en/latest/language_model/index.html",
    ),
    "mfa_mailing_list": ("MFA mailing list", "https://groups.google.com/g/mfa-users"),
    "mfa_github": (
        "MFA GitHub Repo",
        "https://github.com/MontrealCorpusTools/Montreal-Forced-Aligner",
    ),
    "mfa_github_issues": (
        "MFA GitHub Issues",
        "https://github.com/MontrealCorpusTools/Montreal-Forced-Aligner/issues",
    ),
    "memcauliffe.com": ("Michael McAuliffe's blog", "https://memcauliffe.com"),
    "sonderegger": ("Morgan Sonderegger", "http://people.linguistics.mcgill.ca/~morgan/"),
    "wagner": ("Michael Wagner", "https://prosodylab.org/"),
    "coles": ("Arlie Coles", "https://a-coles.github.io/"),
    "chodroff": ("Eleanor Chodroff", "https://www.eleanorchodroff.com/"),
    "dias": ("Danoja Dias", "https://medium.com/breaktheloop"),
    "rutgers_spanish_portuguese": (
        "Rutgers University Department of Spanish and Portuguese",
        "https://span-port.rutgers.edu/",
    ),
    "stengel-eskin": ("Elias Stengel-Eskin", "https://esteng.github.io/"),
    "socolof": ("Michaela Socolof", "https://mcqll.org/people/socolof.michaela/"),
    "mihuc": ("Sarah Mihuc", "https://www.cs.mcgill.ca/~smihuc/"),
    "wsl": ("Windows Subsystem for Linux", "https://docs.microsoft.com/en-us/windows/wsl/install"),
    "kaldi": ("Kaldi", "http://kaldi-asr.org/"),
    "kaldi_github": ("Kaldi GitHub", "https://github.com/kaldi-asr/kaldi"),
    "htk": ("HTK", "http://htk.eng.cam.ac.uk/"),
    "phonetisaurus": ("Phonetisaurus", "https://github.com/AdolfVonKleist/Phonetisaurus"),
    "opengrm_ngram": ("OpenGrm-NGram", "https://www.openfst.org/twiki/bin/view/GRM/NGramLibrary"),
    "openfst": ("OpenFst", "https://www.openfst.org/twiki/bin/view/FST"),
    "baumwelch": ("Baum-Welch", "https://www.opengrm.org/twiki/bin/view/GRM/BaumWelch"),
    "pynini": ("Pynini", "https://www.openfst.org/twiki/bin/view/GRM/Pynini"),
    "prosodylab_aligner": ("Prosodylab-aligner", "http://prosodylab.org/tools/aligner/"),
    "p2fa": (
        "Penn Phonetics Forced Aligner",
        "https://www.ling.upenn.edu/phonetics/old_website_2015/p2fa/",
    ),
    "fave": ("FAVE-align", "https://github.com/JoFrhwld/FAVE/wiki/FAVE-align"),
    "maus": ("MAUS", "http://www.bas.uni-muenchen.de/Bas/BasMAUS.html"),
    "praat": ("Praat", "http://www.fon.hum.uva.nl/praat/"),
    "easy_align": ("EasyAlign", "http://latlcui.unige.ch/phonetique/easyalign.php"),
    "gentle": ("Gentle", "https://lowerquality.com/gentle/"),
    "chodroff_kaldi": ("Kaldi tutorial", "https://eleanorchodroff.com/tutorial/kaldi/index.html"),
    "chodroff_phonetics": (
        "Corpus Phonetics Tutorial",
        "https://eleanorchodroff.com/tutorial/intro.html",
    ),
    "conda_installation": (
        "Conda installation",
        "https://docs.conda.io/projects/conda/en/latest/user-guide/install/index.html",
    ),
    "miniconda": (
        "Miniconda",
        "https://docs.conda.io/en/latest/miniconda.html",
    ),
    "conda_forge": ("Conda Forge", "https://conda-forge.org/"),
    "pydata_sphinx_theme": (
        "Pydata Sphinx Theme",
        "https://pydata-sphinx-theme.readthedocs.io/en/latest/",
    ),
    "mfa_reorg_scripts": (
        "MFA-reorganization-scripts repository",
        "https://github.com/MontrealCorpusTools/MFA-reorganization-scripts",
    ),
    "corpus_creation_scripts": (
        "@mmcauliffe's corpus creation scripts",
        "https://github.com/mmcauliffe/corpus-creation-scripts",
    ),
    "kalpy": (
        "Kalpy",
        "https://kalpy.readthedocs.io/en/latest/",
    ),
    "kalpy_github": (
        "Kalpy GitHub Repo",
        "https://github.com/mmcauliffe/kalpy",
    ),
}

# -----------------------------------------------------------------------------
# Autosummary
# -----------------------------------------------------------------------------


autosummary_generate = True
autodoc_typehints = "none"
# autodoc_typehints_description_target = 'documented'
# autoclass_content = 'both'
autodoc_docstring_signature = True
autodoc_type_aliases = {}

napoleon_preprocess_types = False
napoleon_attr_annotations = False
napoleon_use_param = True
napoleon_use_ivar = True
napoleon_type_aliases = {
    "Labels": "list[str]",
}
typehints_fully_qualified = False
# numpydoc_xref_param_type = True
# numpydoc_show_inherited_class_members = False
numpydoc_show_class_members = False
# -----------------------------------------------------------------------------
# Autodoc
# -----------------------------------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The encoding of source files.
#
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Kalpy"
copyright = f"2023-{date.today().year}, Montreal Corpus Tools"
author = "Montreal Corpus Tools"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "0.5"
# The full version, including alpha/beta/rc tags.
release = "0.5.0"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#
# today = ''
#
# Else, today_fmt is used as the format for a strftime call.
#
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
#
default_role = "code"

# If true, '()' will be appended to :func: etc. cross-reference text.
#
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#
# show_authors = False

# nitpicky = True
nitpick_ignore = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.ico"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "external_links": [
        {
            "url": "https://montreal-forced-aligner.readthedocs.io/",
            "name": "Montreal Forced Aligner",
        },
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/mmcauliffe/kalpy",
            "icon": "fab fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/kalpy-kaldi/",
            "icon": "fas fa-box",
        },
        {
            "name": "PyPI",
            "url": "https://anaconda.org/conda-forge/kalpy",
            "icon": "fas fa-toolbox",
        },
    ],
    "logo": {
        "text": "Kalpy",
        # "image_dark": "logo-dark.svg",
    },
    "analytics": {
        "google_analytics_id": "G-5NGNLY0CWX",
    },
    # "show_nav_level": 1,
    # "navigation_depth": 4,
    # "show_toc_level": 2,
    # "collapse_navigation": True,
}
html_context = {
    "github_user": "mmcauliffe",
    "github_repo": "kalpy",
    "github_version": "main",
    "doc_path": "docs/source",
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
#
# html_title = 'Montreal Forced Aligner v1.0'

# A shorter title for the navigation bar.  Default is the same as html_title.
#
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#
# html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.3/css/fontawesome.min.css",
    "https://montreal-forced-aligner.readthedocs.io/en/latest/_static/css/mfa.css",
]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#
# html_extra_path = []

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
#
# html_last_updated_fmt = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.

html_sidebars = {"**": ["search-field.html", "sidebar-nav-bs.html", "sidebar-ethical-ads.html"]}
# Additional templates that should be rendered to pages, maps page names to
# template names.
#
# html_additional_pages = {}

# If false, no module index is generated.
#
# html_domain_indices = True

# If false, no index is generated.
#
# html_use_index = True

# If true, the index is split into individual pages for each letter.
#
# html_split_index = False

# If true, links to the reST sources are added to the pages.
#
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr', 'zh'
#
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# 'ja' uses this config value.
# 'zh' user can custom change `jieba` dictionary path.
#
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = "Kalpydoc"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).


# The name of an image file (relative to this directory) to place at the top of
# the title page.
#
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#
# latex_use_parts = False

# If true, show page references after internal links.
#
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
#
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
#
# latex_appendices = []

# If false, no module index is generated.
#
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).


# If true, show URL addresses after external links.
#
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = []

# Documents to append as an appendix to all manuals.
#
# texinfo_appendices = []

# If false, no module index is generated.
#
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#
# texinfo_no_detailmenu = False



"""Classes for computing MFCC features"""
from __future__ import annotations

import logging
import pathlib
import typing

import numpy as np

from _kalpy import transform
from _kalpy.matrix import CompressedMatrix, DoubleMatrix, FloatMatrix
from _kalpy.util import (
    BaseDoubleMatrixWriter,
    CompressedMatrixWriter,
    RandomAccessBaseFloatMatrixReader,
)
from kalpy.data import KaldiMapping
from kalpy.feat.data import FeatureArchive
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.cmvn")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class CmvnComputer:
    """
    Class for computing CMVN for features

    Parameters
    ----------
    online: bool

    """

    def __init__(
        self,
        online: bool = False,
    ):
        self.online = online
        self.num_done = 0
        self.num_error = 0

    def compute_cmvn(
        self,
        utterance_list: typing.List[str],
        feature_reader: RandomAccessBaseFloatMatrixReader,
    ) -> np.ndarray:
        """
        Calculate CMVN for a set of utterances

        Parameters
        ----------
        utterance_list: list[str]
            List of utterances to compute CMVN for
        feature_reader: :class:`~_kalpy.util.RandomAccessBaseFloatMatrixReader`
            Reader object for feature file

        Returns
        -------
        :class:`numpy.ndarray`
            Feature matrix for the segment
        """
        cmvn = self.compute_cmvn_for_export(utterance_list, feature_reader)
        return cmvn.numpy()

    def compute_cmvn_from_features(self, features):
        cmvn_stats = DoubleMatrix()
        for i, feats in enumerate(features):
            if i == 0:
                transform.InitCmvnStats(feats.NumCols(), cmvn_stats)
            transform.AccCmvnStats(feats, None, cmvn_stats)
        return cmvn_stats

    def compute_cmvn_for_export(
        self,
        utterance_list: typing.List[str],
        feature_reader: RandomAccessBaseFloatMatrixReader,
    ) -> FloatMatrix:
        """
        Generate MFCCs for exporting to a kaldi archive

        Parameters
        ----------
        utterance_list: list[str]
            List of utterances to compute CMVN for
        feature_reader: :class:`~_kalpy.util.RandomAccessBaseFloatMatrixReader`
            Reader object for feature file

        Returns
        -------
        :class:`_kalpy.matrix.DoubleMatrix`
            Feature matrix for the segment
        """
        cmvn_stats, num_done, num_error = transform.calculate_cmvn(utterance_list, feature_reader)
        self.num_done += num_done
        self.num_error += num_error
        return cmvn_stats

    def export_cmvn(
        self,
        file_name: typing.Union[pathlib.Path, str],
        feature_archive: FeatureArchive,
        spk2utt: KaldiMapping,
        write_scp: bool = False,
        compress: bool = False,
    ) -> None:
        """
        Export features to a kaldi archive file (i.e., cmvn.ark)

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Archive file path to export to
        feature_archive: :class:`~kalpy.feat.data.FeatureArchive`
            Archive of features
        spk2utt: :class:`~kalpy.data.KaldiMapping`
            Mapping of speaker ids to utterance ids
        write_scp: bool
            Flag for whether an SCP file should be generated as well
        compress: bool
            Flag for whether to export features as a compressed archive
        """
        write_specifier = generate_write_specifier(file_name, write_scp)
        logger.debug(f"Writing to: {write_specifier}")
        if compress:
            writer = CompressedMatrixWriter(write_specifier)
        else:
            writer = BaseDoubleMatrixWriter(write_specifier)
        feat_reader = feature_archive.archive.random_reader
        for key, utterance_list in spk2utt.items():
            logger.info(f"Processing speaker {key}: {len(utterance_list)} utterances")
            feats = self.compute_cmvn_for_export(utterance_list, feat_reader)
            if compress:
                feats = CompressedMatrix(feats)
            writer.Write(str(key), feats)
        writer.Close()



import os

import numpy as np
import pytest

from _kalpy import transform
from _kalpy.feat import RandomAccessWaveReader
from _kalpy.matrix import CompressedMatrix
from _kalpy.util import (
    BaseFloatMatrixWriter,
    CompressedMatrixWriter,
    RandomAccessBaseDoubleMatrixReader,
    RandomAccessBaseFloatMatrixReader,
)
from kalpy.data import KaldiMapping, MatrixArchive, Segment
from kalpy.decoder.training_graphs import TrainingGraphCompiler
from kalpy.feat.cmvn import CmvnComputer
from kalpy.feat.data import FeatureArchive
from kalpy.feat.fmllr import FmllrComputer
from kalpy.feat.mfcc import MfccComputer
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.gmm.align import GmmAligner
from kalpy.gmm.data import AlignmentArchive, LatticeArchive
from kalpy.utils import generate_read_specifier, generate_write_specifier


@pytest.mark.order(1)
def test_wave(wav_path, reference_dir):
    ref_wav_scp = reference_dir.joinpath("wav.scp")
    wav_rspecifier = generate_read_specifier(ref_wav_scp)
    wave_reader = RandomAccessWaveReader(wav_rspecifier)
    kaldi_wave = wave_reader.Value("1-1").Data().numpy()[0, :]
    segment = Segment(wav_path)
    kalpy_wave = segment.kaldi_wave

    np.testing.assert_allclose(kaldi_wave, kalpy_wave)


@pytest.mark.order(1)
def test_generate_mfcc(wav_path):
    feature_generator = MfccComputer(snip_edges=False)
    mfccs = feature_generator.compute_mfccs(Segment(wav_path))
    assert mfccs.shape[0] == 2672
    assert mfccs.shape[1] == 13
    mfccs = feature_generator.compute_mfccs(Segment(wav_path, 1, 2, 0))
    assert mfccs.shape[0] == 100
    assert mfccs.shape[1] == 13


@pytest.mark.order(1)
def test_export_mfcc(wav_path, mono_temp_dir):
    output_file_name = mono_temp_dir.joinpath("mfccs.ark")
    feature_generator = MfccComputer(snip_edges=False)
    segments = {"1-1": Segment(wav_path)}
    feature_generator.export_feats(output_file_name, segments.items())
    assert output_file_name.exists()

    archive = FeatureArchive(output_file_name)
    for utt, mfccs in archive:
        mfccs = mfccs.numpy()
        assert utt == "1-1"
        assert mfccs.shape[0] == 2672
        assert mfccs.shape[1] == 13
    mfccs = archive["1-1"].numpy()
    print(mfccs)
    print(mfccs.shape)
    assert mfccs.shape[0] == 2672
    assert mfccs.shape[1] == 13
    archive.close()
    os.remove(output_file_name)
    feature_generator.export_feats(
        output_file_name, segments.items(), write_scp=True, compress=False
    )
    assert output_file_name.with_suffix(".scp").exists()

    archive = FeatureArchive(output_file_name.with_suffix(".scp"))
    for utt, mfccs in archive:
        mfccs = mfccs.numpy()
        assert utt == "1-1"
        assert mfccs.shape[0] == 2672
        assert mfccs.shape[1] == 13
    mfccs = archive["1-1"].numpy()
    assert mfccs.shape[0] == 2672
    assert mfccs.shape[1] == 13


@pytest.mark.order(2)
def test_cmvn(mono_temp_dir):
    feature_file_name = mono_temp_dir.joinpath("mfccs.ark")
    output_file_name = mono_temp_dir.joinpath("cmvn.ark")
    feature_generator = CmvnComputer(online=False)
    spk2utt = KaldiMapping()
    spk2utt["1"] = ["1-1"]
    feature_archive = FeatureArchive(feature_file_name)
    feature_generator.export_cmvn(output_file_name, feature_archive, spk2utt)
    assert output_file_name.exists()


@pytest.mark.order(1)
def test_export_mfcc_sat(wav_path, sat_temp_dir, reference_dir, reference_mfcc_path):
    output_file_name = sat_temp_dir.joinpath("mfccs.ark")
    feature_generator = MfccComputer(snip_edges=False, dither=0, use_energy=False)
    segments = {"1-1": Segment(wav_path, begin=0.0, end=26.72325)}
    feature_generator.export_feats(output_file_name, segments.items())
    assert output_file_name.exists()

    archive = FeatureArchive(output_file_name)
    for utt, mfccs in archive:
        mfccs = mfccs.numpy()
        assert utt == "1-1"
        assert mfccs.shape[0] == 2672
        assert mfccs.shape[1] == 13
    mfccs = archive["1-1"].numpy()
    print(mfccs)
    print(mfccs.shape)
    assert mfccs.shape[0] == 2672
    assert mfccs.shape[1] == 13

    archive.close()
    os.remove(output_file_name)
    feature_generator.export_feats(
        output_file_name, segments.items(), write_scp=True, compress=False
    )
    assert output_file_name.with_suffix(".scp").exists()

    archive = FeatureArchive(output_file_name.with_suffix(".scp"))
    ref_archive = FeatureArchive(reference_mfcc_path)
    for utt, mfccs in archive:
        mfccs = mfccs.numpy()
        assert utt == "1-1"
        assert mfccs.shape[0] == 2672
        assert mfccs.shape[1] == 13
        ref_mfccs = ref_archive["1-1"]
        np.testing.assert_allclose(mfccs, ref_mfccs.numpy())
    mfccs = archive["1-1"].numpy()
    assert mfccs.shape[0] == 2672
    assert mfccs.shape[1] == 13


@pytest.mark.order(2)
def test_cmvn_sat(
    sat_temp_dir,
    reference_dir,
    reference_mfcc_path,
    reference_cmvn_path,
    reference_final_features_path,
):
    feature_file_name = sat_temp_dir.joinpath("mfccs.ark")
    output_file_name = sat_temp_dir.joinpath("cmvn.ark")
    feature_generator = CmvnComputer(online=False)
    spk2utt = KaldiMapping()
    spk2utt["1"] = ["1-1"]
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    feature_archive = FeatureArchive(feature_file_name)
    feature_generator.export_cmvn(output_file_name, feature_archive, spk2utt, write_scp=True)
    assert output_file_name.exists()
    cmvn_read_specifier = generate_read_specifier(output_file_name)
    cmvn_reader = RandomAccessBaseDoubleMatrixReader(cmvn_read_specifier)
    ref_cmvn_read_specifier = generate_read_specifier(reference_cmvn_path)
    ref_cmvn_reader = RandomAccessBaseDoubleMatrixReader(ref_cmvn_read_specifier)
    cmvn = cmvn_reader.Value("1")
    ref_cmvn = ref_cmvn_reader.Value("1")
    np.testing.assert_allclose(cmvn.numpy(), ref_cmvn.numpy())
    final_archive = FeatureArchive(
        feature_file_name,
        utt2spk=utt2spk,
        cmvn_file_name=output_file_name,
    )
    feat_archive = FeatureArchive(
        feature_file_name,
    )
    ref_final_features_archive = FeatureArchive(reference_final_features_path)
    feats = feat_archive["1-1"]
    cmvned_feats = transform.apply_cmvn(feats, cmvn)
    ref_cmvned_feats = ref_final_features_archive["1-1"]
    np.testing.assert_allclose(cmvned_feats.numpy(), ref_cmvned_feats.numpy())
    temp_ark_path = sat_temp_dir.joinpath("final_features.ark")
    temp_scp_path = sat_temp_dir.joinpath("final_features.scp")
    write_specifier = generate_write_specifier(temp_ark_path, write_scp=True)
    feature_writer = BaseFloatMatrixWriter(write_specifier)
    np.testing.assert_allclose(
        final_archive["1-1"].numpy(), ref_final_features_archive["1-1"].numpy()
    )
    for utt_id, mfccs in final_archive:
        feature_writer.Write(utt_id, mfccs)
    feature_writer.Close()
    feature_archive = FeatureArchive(temp_scp_path)
    np.testing.assert_allclose(
        feature_archive["1-1"].numpy(), ref_final_features_archive["1-1"].numpy()
    )
    cmvn_reader.Close()
    ref_cmvn_reader.Close()


@pytest.mark.order(4)
def test_fmllr_sat(
    sat_tree_path,
    sat_model_path,
    sat_align_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    reference_trans_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones)
    lc.load_pronunciations(sat_dictionary_path)
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        splices=True,
    )
    alignment_archive = AlignmentArchive(sat_temp_dir.joinpath("ali.ark"))
    output_file_name = sat_temp_dir.joinpath("trans.ark")
    spk2utt = KaldiMapping()
    spk2utt["1"] = ["1-1"]
    fmllr_computer = FmllrComputer(
        sat_align_model_path, sat_model_path, silence_phones=lc.silence_symbols, spk2utt=spk2utt
    )
    fmllr_computer.export_transforms(output_file_name, feature_archive, alignment_archive)
    assert output_file_name.exists()
    feature_archive.close()
    alignment_archive.close()
    trans_read_specifier = generate_read_specifier(output_file_name)
    trans_reader = RandomAccessBaseFloatMatrixReader(trans_read_specifier)
    ref_trans_read_specifier = generate_read_specifier(reference_trans_path)
    ref_trans_reader = RandomAccessBaseFloatMatrixReader(ref_trans_read_specifier)
    trans = trans_reader.Value("1")
    ref_trans = ref_trans_reader.Value("1")
    np.testing.assert_allclose(trans.numpy(), ref_trans.numpy())


@pytest.mark.order(4)
def test_fmllr_sat_no_two_model(
    sat_tree_path,
    sat_align_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    reference_trans_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones)
    lc.load_pronunciations(sat_dictionary_path)
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        splices=True,
    )
    alignment_archive = AlignmentArchive(sat_temp_dir.joinpath("ali.ark"))
    output_file_name = sat_temp_dir.joinpath("trans_no_two_model.ark")
    spk2utt = KaldiMapping()
    spk2utt["1"] = ["1-1"]
    fmllr_computer = FmllrComputer(
        sat_align_model_path,
        sat_align_model_path,
        silence_phones=lc.silence_symbols,
        spk2utt=spk2utt,
    )
    fmllr_computer.export_transforms(output_file_name, feature_archive, alignment_archive)
    assert output_file_name.exists()
    feature_archive.close()
    alignment_archive.close()


@pytest.mark.order(4)
def test_fmllr_decode_sat(
    sat_tree_path,
    sat_model_path,
    sat_align_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones)
    lc.load_pronunciations(sat_dictionary_path)
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        splices=True,
    )
    alignment_archive = LatticeArchive(sat_temp_dir.joinpath("lat.ark"), determinized=False)
    output_file_name = sat_temp_dir.joinpath("trans_decode.ark")
    spk2utt = KaldiMapping()
    spk2utt["1"] = ["1-1"]
    fmllr_computer = FmllrComputer(
        sat_align_model_path, sat_model_path, silence_phones=lc.silence_symbols, spk2utt=spk2utt
    )
    fmllr_computer.export_transforms(output_file_name, feature_archive, alignment_archive)
    assert output_file_name.exists()


@pytest.mark.order(6)
def test_fmllr_compose(
    sat_tree_path,
    sat_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    reference_trans_path,
    reference_trans_compose_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones)
    lc.load_pronunciations(sat_dictionary_path)
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        transform_file_name=reference_trans_path,
        splices=True,
    )
    previous_transform_archive = MatrixArchive(reference_trans_path)
    alignment_archive = AlignmentArchive(sat_temp_dir.joinpath("ali_second_pass.ark"))
    output_file_name = sat_temp_dir.joinpath("trans_second_pass.ark")
    spk2utt = KaldiMapping()
    spk2utt["1"] = ["1-1"]
    fmllr_computer = FmllrComputer(
        sat_model_path, sat_model_path, silence_phones=lc.silence_symbols, spk2utt=spk2utt
    )
    fmllr_computer.export_transforms(
        output_file_name,
        feature_archive,
        alignment_archive,
        previous_transform_archive=previous_transform_archive,
        write_scp=True,
    )
    assert output_file_name.exists()
    feature_archive.close()
    alignment_archive.close()
    trans_read_specifier = generate_read_specifier(output_file_name)
    trans_reader = RandomAccessBaseFloatMatrixReader(trans_read_specifier)
    ref_trans_composed_read_specifier = generate_read_specifier(reference_trans_compose_path)
    ref_trans_composed_reader = RandomAccessBaseFloatMatrixReader(
        ref_trans_composed_read_specifier
    )
    ref_trans_read_specifier = generate_read_specifier(reference_trans_path)
    ref_trans_reader = RandomAccessBaseFloatMatrixReader(ref_trans_read_specifier)
    trans = trans_reader.Value("1")
    ref_trans = ref_trans_reader.Value("1")
    ref_trans_composed = ref_trans_composed_reader.Value("1")
    np.testing.assert_allclose(trans.numpy(), ref_trans_composed.numpy())
    np.testing.assert_raises(
        AssertionError, np.testing.assert_allclose, trans.numpy(), ref_trans.numpy()
    )



from __future__ import annotations

import logging
import pathlib
import typing
from contextlib import contextmanager

from _kalpy.util import Input, Output


def generate_read_specifier(file_name: typing.Union[str, pathlib.Path], sorted=True) -> str:
    file_name = str(file_name)
    if sorted:
        read_identifier = "ark,s,cs"
    else:
        read_identifier = "ark"
    if file_name.endswith(".scp"):
        if sorted:
            read_identifier = "scp,s,cs"
        else:
            read_identifier = "scp"
    return f"{read_identifier}:{file_name}"


def generate_write_specifier(
    file_name: typing.Union[str, pathlib.Path], write_scp: bool = False, text: bool = False
) -> str:
    file_name = str(file_name)
    if not file_name.endswith(".ark"):
        file_name += ".ark"
    ark_prefix = "ark"
    if text:
        ark_prefix += ",t"
    if write_scp:
        return f"{ark_prefix},scp:{file_name},{file_name.replace('.ark', '.scp')}"
    return f"{ark_prefix}:{file_name}"


def write_kaldi_object(obj, path, binary=True):
    ko = Output(str(path), binary)
    obj.Write(ko.Stream(), binary)
    ko.Close()


def read_kaldi_object(obj_type, path, binary=True):
    ki = Input()
    ki.Open(str(path), True)
    obj = obj_type()
    obj.Read(ki.Stream(), binary)
    ki.Close()
    return obj


@contextmanager
def kalpy_logger(log_name: str, log_path: typing.Union[pathlib.Path, str]) -> logging.Logger:
    kalpy_logging = logging.getLogger(log_name)
    file_handler = logging.FileHandler(log_path, encoding="utf8")
    file_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)
    try:
        kalpy_logging.addHandler(file_handler)
        yield kalpy_logging
    finally:
        file_handler.close()
        kalpy_logging.removeHandler(file_handler)


def get_kalpy_version() -> str:
    """
    Get the current Kalpy version

    Returns
    -------
    str
        Kalpy version
    """
    try:
        from ._version import version as __version__  # noqa
    except ImportError:
        __version__ = "0.2.0"
    return __version__



"""Classes for GMM alignment"""
from __future__ import annotations

import logging
import pathlib
import typing

from _kalpy.decoder import LatticeFasterDecoder, LatticeFasterDecoderConfig
from _kalpy.fstext import ConstFst, GetLinearSymbolSequence
from _kalpy.gmm import DecodableAmDiagGmmScaled, gmm_rescore_lattice
from _kalpy.lat import (
    CompactLattice,
    CompactLatticeWriter,
    DeterminizeLatticePhonePrunedWrapper,
    GetPerFrameAcousticCosts,
    LatticeWriter,
    lattice_to_post,
)
from _kalpy.matrix import FloatMatrix
from _kalpy.util import Int32VectorWriter
from kalpy.feat.data import FeatureArchive
from kalpy.gmm.data import Alignment, LatticeArchive
from kalpy.gmm.utils import read_gmm_model
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.decode")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class GmmDecoder:
    def __init__(
        self,
        acoustic_model_path: typing.Union[pathlib.Path, str],
        hclg_fst: ConstFst,
        acoustic_scale: float = 0.1,
        beam: float = 16.0,
        lattice_beam: float = 10.0,
        max_active: int = 7000,
        min_active: int = 200,
        prune_interval: int = 25,
        determinize_lattice: bool = True,
        beam_delta: float = 0.5,
        hash_ratio: float = 2.0,
        prune_scale: float = 0.1,
        allow_partial: bool = True,
        fast: bool = False,
    ):
        self.acoustic_model_path = acoustic_model_path
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.hclg_fst = hclg_fst
        self.acoustic_scale = acoustic_scale
        self.allow_partial = allow_partial
        self.beam = beam
        self.lattice_beam = lattice_beam
        self.max_active = max_active
        self.min_active = min_active
        self.prune_interval = prune_interval
        self.determinize_lattice = determinize_lattice
        self.beam_delta = beam_delta
        self.hash_ratio = hash_ratio
        self.prune_scale = prune_scale
        self.fast = fast

        self.config = LatticeFasterDecoderConfig()
        self.config.beam = beam
        self.config.lattice_beam = lattice_beam
        self.config.max_active = max_active
        self.config.min_active = min_active
        self.config.prune_interval = prune_interval
        self.config.determinize_lattice = determinize_lattice
        self.config.beam_delta = beam_delta
        self.config.hash_ratio = hash_ratio
        self.config.prune_scale = prune_scale
        self.num_done = 0
        self.num_error = 0
        self.num_retry = 0
        self.total_likelihood = 0
        self.total_frames = 0

    def boost_silence(self, silence_weight: float, silence_phones: typing.List[int]):
        if silence_weight != 1.0:
            self.acoustic_model.boost_silence(
                self.transition_model, silence_phones, silence_weight
            )

    def decode_utterance(
        self, features: FloatMatrix, utterance_id: str = None
    ) -> typing.Optional[Alignment]:
        decodable = DecodableAmDiagGmmScaled(
            self.acoustic_model, self.transition_model, features, self.acoustic_scale
        )

        d = LatticeFasterDecoder(self.hclg_fst, self.config)
        ans = d.Decode(decodable)
        if not ans:
            logger.warning(f"Did not successfully decode {utterance_id}")
            self.num_error += 1
            return None

        ans = d.ReachedFinal()
        if not ans:
            if self.allow_partial:
                logger.warning(
                    f"Outputting partial output for utterance {utterance_id} "
                    "since no final-state reached"
                )
            else:
                logger.warning(
                    f"Not producing output for utterance {utterance_id} "
                    f"since no final-state reached and allow_partial==False"
                )

        ans, decoded = d.GetBestPath()
        if not ans:
            logger.error(f"Failed to get traceback for utterance {utterance_id}")
            return None
        if decoded.NumStates() == 0:
            logger.warning(f"Error getting best path from decoder for utterance {utterance_id}")
        alignment, words, weight = GetLinearSymbolSequence(decoded)
        likelihood = -(weight.Value1() + weight.Value2()) / self.acoustic_scale
        if self.fast:
            return Alignment(utterance_id, alignment, words, likelihood)

        self.num_done += 1
        self.total_likelihood += likelihood
        self.total_frames += len(alignment)
        per_frame_log_likelihoods = GetPerFrameAcousticCosts(decoded)
        per_frame_log_likelihoods.Scale(-1 / self.acoustic_scale)
        ans, lat = d.GetRawLattice()
        lat.Connect()
        if self.config.determinize_lattice:
            clat = CompactLattice()
            ans = DeterminizeLatticePhonePrunedWrapper(
                self.transition_model,
                lat,
                self.config.lattice_beam,
                clat,
                self.config.det_opts,
            )
            if not ans:
                logger.warning(
                    f"Determinization finished earlier than the beam for utterance {utterance_id}"
                )
            if self.acoustic_scale != 0.0:
                clat.ScaleLattice(acoustic_scale=self.acoustic_scale)
            data = Alignment(
                utterance_id,
                alignment,
                words,
                likelihood,
                per_frame_log_likelihoods,
                lattice=clat,
            )
        else:
            if self.acoustic_scale != 0.0:
                lat.ScaleLattice(acoustic_scale=self.acoustic_scale)
            data = Alignment(
                utterance_id,
                alignment,
                words,
                likelihood,
                per_frame_log_likelihoods,
                lattice=lat,
            )
        logger.info(
            f"Log-like per frame for utterance {utterance_id} is "
            f"{likelihood/len(alignment)} over {len(alignment)} frames."
        )
        logger.debug(
            f"Cost for utterance {utterance_id} is " f"{weight.Value1()} + {weight.Value2()}"
        )
        return data

    def decode_utterances(self, feature_archive: FeatureArchive) -> typing.Generator[Alignment]:
        self.num_done = 0
        self.num_error = 0
        self.num_retry = 0
        self.total_likelihood = 0
        self.total_frames = 0
        for (utterance_id, feats) in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utterance_id} due to zero-length features")
                continue
            yield self.decode_utterance(feats, utterance_id)
        logger.info(
            f"Overall log-likelihood per frame is {self.total_likelihood / self.total_frames} "
            f"over {self.total_frames} frames."
        )
        logger.info(
            f"Retried {self.num_retry} out of {self.num_done + self.num_error} utterances."
        )
        logger.info(f"Done {self.num_done}, errors on {self.num_error}")

    def export_lattices(
        self,
        file_name: typing.Union[str, pathlib.Path],
        feature_archive: FeatureArchive,
        write_scp: bool = False,
        alignment_file_name: typing.Union[str, pathlib.Path] = None,
        word_file_name: typing.Union[str, pathlib.Path] = None,
        callback: typing.Callable = None,
    ):
        write_specifier = generate_write_specifier(file_name, write_scp)
        alignment_writer = None
        if alignment_file_name:
            alignment_write_specifier = generate_write_specifier(alignment_file_name, write_scp)
            alignment_writer = Int32VectorWriter(alignment_write_specifier)
        word_writer = None
        if word_file_name:
            word_write_specifier = generate_write_specifier(word_file_name, write_scp)
            word_writer = Int32VectorWriter(word_write_specifier)
        if self.config.determinize_lattice:
            writer = CompactLatticeWriter(write_specifier)
        else:
            writer = LatticeWriter(write_specifier)
        for transcription in self.decode_utterances(feature_archive):
            if transcription is None:
                continue
            if callback:
                callback((transcription.utterance_id, transcription.likelihood))
            writer.Write(str(transcription.utterance_id), transcription.lattice)
            if alignment_writer is not None:
                alignment_writer.Write(str(transcription.utterance_id), transcription.alignment)
            if word_writer is not None:
                word_writer.Write(str(transcription.utterance_id), transcription.words)
        writer.Close()
        if alignment_writer is not None:
            alignment_writer.Close()
        if word_writer is not None:
            word_writer.Close()


class GmmRescorer:
    def __init__(
        self,
        acoustic_model_path: typing.Union[pathlib.Path, str],
        acoustic_scale: float = 0.1,
        lattice_beam: float = 6.0,
    ):
        self.acoustic_model_path = acoustic_model_path
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.acoustic_scale = acoustic_scale
        self.lattice_beam = lattice_beam
        self.num_done = 0
        self.num_error = 0
        self.config = LatticeFasterDecoderConfig()

    def rescore_utterance(
        self, lattice: CompactLattice, feats: FloatMatrix, utterance_id: str = None
    ):

        try:
            ans, lattice = gmm_rescore_lattice(
                lattice, feats, self.acoustic_model, self.transition_model
            )
            if not ans:
                raise Exception(f"Error in rescoring {utterance_id}")

            clat = lattice_to_post(lattice, self.lattice_beam, self.acoustic_scale)
            self.num_done += 1
        except Exception:
            self.num_error += 1
            logger.warning(f"Error in rescoring {utterance_id}")
            raise
        return clat

    def rescore_utterances(
        self, lattice_archive: LatticeArchive, feature_archive: FeatureArchive
    ) -> typing.Generator[CompactLattice]:
        for (utterance_id, lattice) in lattice_archive:
            feats = feature_archive[utterance_id]
            yield utterance_id, self.rescore_utterance(lattice, feats, utterance_id)

    def export_lattices(
        self,
        file_name: typing.Union[str, pathlib.Path],
        lattice_archive: LatticeArchive,
        feature_archive: FeatureArchive,
        write_scp: bool = False,
        callback: typing.Callable = None,
    ):
        write_specifier = generate_write_specifier(file_name, write_scp)
        writer = CompactLatticeWriter(write_specifier)
        for utterance_id, lattice in self.rescore_utterances(lattice_archive, feature_archive):
            if callback:
                callback(utterance_id)
            writer.Write(str(utterance_id), lattice)
        writer.Close()



import pathlib
import typing

import numpy as np

from _kalpy.ivector import Plda, ivector_normalize_length, ivector_subtract_mean
from _kalpy.matrix import DoubleVector, FloatVector
from kalpy.utils import read_kaldi_object
from kalpy.ivector.data import IvectorArchive


class PldaScorer:
    def __init__(
        self,
        plda_path: typing.Union[str, pathlib.Path],
        normalize_length: bool = True,
        simple_length_norm: bool = True,
    ):
        self.plda_path = str(plda_path)
        self.plda: Plda = read_kaldi_object(Plda, self.plda_path)
        self.normalize_length = normalize_length
        self.simple_length_norm = simple_length_norm
        self.speaker_ids = None
        self.speaker_ivectors = None
        self.num_speaker_examples = None

    def load_speaker_ivectors(self, speaker_archive_path, num_utts_path=None):
        ivector_archive = IvectorArchive(
            speaker_archive_path, num_utterances_file_name=num_utts_path
        )
        speaker_ivectors = []
        self.speaker_ids = []
        self.num_speaker_examples = []
        for speaker_id, ivector, utts in ivector_archive:
            self.speaker_ids.append(speaker_id)
            self.num_speaker_examples.append(utts)
            if self.normalize_length:
                ivector_normalize_length(ivector)
            speaker_ivectors.append(DoubleVector(ivector))
        ivector_subtract_mean(speaker_ivectors,normalize=self.normalize_length)
        self.speaker_ivectors = self.plda.transform_ivectors(speaker_ivectors, self.num_speaker_examples)

    def transform_ivector(self, ivector: np.ndarray, num_examples: int = 1):
        return self.plda.transform_ivector(ivector, num_examples)

    def transform_ivectors(self, ivectors: np.ndarray, num_examples: np.ndarray = None):
        if num_examples is None:
            num_examples = np.ones((ivectors.shape[0]))
        return self.plda.transform_ivectors(ivectors, num_examples)

    def score_ivectors(
        self,
        speaker_ivector: typing.Union[np.ndarray, FloatVector, DoubleVector],
        utterance_ivector: typing.Union[np.ndarray, FloatVector, DoubleVector],
        num_speaker_examples: int = 1,
    ):
        score = self.plda.LogLikelihoodRatio(
            speaker_ivector, num_speaker_examples, utterance_ivector
        )
        return score

    def classify_speaker(
        self,
        utterance_ivector: typing.Union[np.ndarray, FloatVector, DoubleVector],
    ):
        if self.num_speaker_examples is None:
            self.num_speaker_examples = [1 for _ in (self.speaker_ivectors.shape[0])]
        if isinstance(utterance_ivector, np.ndarray):
            utterance_ivector = DoubleVector()
            utterance_ivector.from_numpy(utterance_ivector)
        ind, score = self.plda.classify_utterance(utterance_ivector, self.speaker_ivectors, self.num_speaker_examples)
        speaker = self.speaker_ids[ind]
        return speaker, score



import numpy as np

from _kalpy.matrix import FloatMatrix


def test_casting():
    t = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype="float32")
    mat = FloatMatrix()
    mat.from_numpy(t)
    print(mat.Stride())
    assert mat.NumRows() == 2
    assert mat.NumCols() == 3
    assert np.all(mat.numpy() == t)



"""Classes for storing graph archives"""
from __future__ import annotations

import os.path
import pathlib
import typing

from _kalpy.fstext import RandomAccessVectorFstReader, SequentialVectorFstReader, VectorFst
from kalpy.utils import generate_read_specifier


class FstArchive:
    """
    Class for reading an archive or SCP of FSTs

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(self, file_name: typing.Union[pathlib.Path, str]):
        if not os.path.exists(file_name):
            raise OSError(f"Specified file does not exist: {file_name}")
        self.file_name = str(file_name)
        self.read_specifier = generate_read_specifier(file_name)
        self.random_reader = RandomAccessVectorFstReader(self.read_specifier)

    @property
    def sequential_reader(self) -> SequentialVectorFstReader:
        return SequentialVectorFstReader(self.read_specifier)

    def __iter__(self) -> typing.Generator[typing.Tuple[str, VectorFst]]:
        """Iterate over the utterance FSTs in the archive"""
        reader = self.sequential_reader
        try:
            while not reader.Done():
                utt = reader.Key()
                decode_fst = VectorFst(reader.Value())
                reader.FreeCurrent()
                yield utt, decode_fst
                reader.Next()
        finally:
            reader.Close()

    def __del__(self):
        if self.random_reader.IsOpen():
            self.random_reader.Close()

    def __getitem__(self, item: str) -> VectorFst:
        """Get FST for a particular key from the archive file"""
        item = str(item)
        if not self.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.file_name}")
        return self.random_reader.Value(item)



"""Utility functions for working with fstext"""
import typing

import pynini

from _kalpy.fstext import ConstFst, VectorFst


def kaldi_to_pynini(fst: typing.Union[VectorFst, ConstFst]) -> pynini.Fst:
    """
    Converts an FST from Kaldi to pynini

    Parameters
    ----------
    fst: :class:`_kalpy.fstext.VectorFst`
        FST from Kaldi

    Returns
    -------
    :class:`~pynini.Fst`
        FST for use in pynini
    """
    return pynini.Fst.read_from_string(fst.write_to_string())


def pynini_to_kaldi(fst: pynini.Fst) -> VectorFst:
    """
    Converts an FST from pynini to Kaldi

    Parameters
    ----------
    fst: :class:`~pynini.Fst`
        FST from pynini

    Returns
    -------
    :class:`_kalpy.fstext.VectorFst`
        VectorFst for use in Kaldi
    """
    return VectorFst.from_string(fst.write_to_string())


def pynini_to_kaldi_const(fst: pynini.Fst) -> ConstFst:
    """
    Converts an FST from pynini to Kaldi ConstFst

    Parameters
    ----------
    fst: :class:`~pynini.Fst`
        FST from pynini

    Returns
    -------
    :class:`_kalpy.fstext.ConstFst`
        ConstFst for use in Kaldi
    """
    return ConstFst.from_string(fst.write_to_string())



"""Classes for working with lexicons"""
from __future__ import annotations

import collections
import math
import pathlib
import re
import typing
from dataclasses import dataclass

import pynini
import pywrapfst

from _kalpy.fstext import VectorFst
from _kalpy.lat import WordAlignLatticeLexiconInfo
from kalpy.exceptions import LexiconError, PhonesToPronunciationsError
from kalpy.gmm.data import CtmInterval, HierarchicalCtm, WordCtmInterval


@dataclass
class Pronunciation:
    """
    Data class for storing information about a particular pronunciation
    """

    orthography: str
    pronunciation: str
    probability: typing.Optional[float] = None
    silence_after_probability: typing.Optional[float] = None
    silence_before_correction: typing.Optional[float] = None
    non_silence_before_correction: typing.Optional[float] = None
    disambiguation: typing.Optional[int] = None


def parse_dictionary_file(
    path: typing.Union[pathlib.Path, str],
) -> typing.Generator[Pronunciation]:
    """
    Parses a lexicon file and yields parsed pronunciation lines

    Parameters
    ----------
    path: :class:`~pathlib.Path` or str
        Path to lexicon file

    Yields
    ------
    str
        :class:`~kalpy.fstext.lexicon.Pronunciation`
    """
    prob_pattern = re.compile(r"\b\d+\.\d+\b")
    found_set = set()
    with open(path, encoding="utf8") as f:
        for i, line in enumerate(f):
            line = line.strip()
            if not line:
                continue
            line = line.split()
            if len(line) <= 1:
                raise Exception(
                    f'Error parsing line {i} of {path}: "{line}" did not have a pronunciation'
                )
            word = line.pop(0)
            prob = None
            silence_after_prob = None
            silence_before_correction = None
            non_silence_before_correction = None
            if prob_pattern.match(line[0]):
                prob = float(line.pop(0))
                if prob_pattern.match(line[0]):
                    silence_after_prob = float(line.pop(0))
                    if prob_pattern.match(line[0]):
                        silence_before_correction = float(line.pop(0))
                        if prob_pattern.match(line[0]):
                            non_silence_before_correction = float(line.pop(0))
            pron = " ".join(line)
            if (word, pron) in found_set:
                continue
            found_set.add((word, pron))
            yield Pronunciation(
                word,
                pron,
                prob,
                silence_after_prob,
                silence_before_correction,
                non_silence_before_correction,
                None,
            )


class LexiconCompiler:
    """
    Class for compiling pronunciation dictionary files to lexicon FSTs

    Parameters
    ----------
    disambiguation: bool
        Flag for compiling a disambiguated lexicon for decoding instead of alignment
    silence_probability: float
        Probability of silence following words
    initial_silence_probability: float
        Probability of silence at the beginning of utterances
    final_silence_correction: float
        Correction factor for utterances ending in silence
    final_non_silence_correction: float
        Correction factor for utterances not ending in silence
    silence_word: str
        Word symbol to use for silence
    oov_word: str
        Word symbol to use for out of vocabulary items
    silence_phone: str
        Phone symbol to use for silence
    oov_phone: str
        Phone symbol to use for out of vocabulary items
    position_dependent_phones: bool
        Flag for using phones based on word position,
        i.e. "AA_S" (for words with a pronunciation of only "AA"),
        "AA_B" ("AA" at the beginning of a word),
        "AA_I" ("AA" internal to  a word),
        "AA_E" ("AA" at the end of a word)
        instead of using a single "AA" symbol across all word positions
    ignore_case: bool
        Flag for whether word orthographies should be transformed to lower case

    Attributes
    ----------
    word_table: :class:`pywrapfst.SymbolTable`
        Word symbol table
    phone_table: :class:`pywrapfst.SymbolTable`
        Phone symbol table
    pronunciations: list[:class:`~kalpy.fstext.lexicon.Pronunciation`]
        List of pronunciations loaded from dictionary file
    """

    use_g2p = False

    def __init__(
        self,
        disambiguation: bool = False,
        silence_probability: float = 0.5,
        initial_silence_probability: float = 0.5,
        final_silence_correction: typing.Optional[float] = None,
        final_non_silence_correction: typing.Optional[float] = None,
        silence_word: str = "<eps>",
        oov_word: str = "<unk>",
        silence_phone: str = "sil",
        oov_phone: str = "spn",
        position_dependent_phones: bool = False,
        ignore_case: bool = True,
        phones: typing.Collection[str] = None,
        word_begin_label: str = "#1",
        word_end_label: str = "#2",
    ):
        self.disambiguation = disambiguation
        self.silence_disambiguation_symbol = "<eps>"
        self.silence_probability = silence_probability
        self.initial_silence_probability = initial_silence_probability
        self.final_silence_correction = final_silence_correction
        self.final_non_silence_correction = final_non_silence_correction
        self.silence_word = silence_word
        self.oov_word = oov_word
        self.silence_phone = silence_phone
        self.oov_phone = oov_phone
        self.max_disambiguation_symbol = 1
        self.position_dependent_phones = position_dependent_phones
        self.ignore_case = ignore_case
        self.word_table = pywrapfst.SymbolTable()
        self.word_table.add_symbol(silence_word)
        self.word_table.add_symbol(oov_word)
        self.phone_table = pywrapfst.SymbolTable()
        self.phone_table.add_symbol("<eps>")
        self.phone_table.add_symbol(silence_phone)
        if self.position_dependent_phones:
            for pos in ["_S", "_B", "_E", "_I"]:
                self.phone_table.add_symbol(silence_phone + pos)
        self.phone_table.add_symbol(oov_phone)
        if self.position_dependent_phones:
            for pos in ["_S", "_B", "_E", "_I"]:
                self.phone_table.add_symbol(oov_phone + pos)
        if phones is not None:
            for p in sorted(phones):
                if self.position_dependent_phones:
                    for pos in ["_S", "_B", "_E", "_I"]:
                        self.phone_table.add_symbol(p + pos)
                else:
                    self.phone_table.add_symbol(p)
        self.pronunciations: typing.List[Pronunciation] = []
        self._cached_pronunciations: typing.Set[typing.Tuple[str, str]] = set()
        self._fst = None
        self._align_fst = None
        self._align_lexicon = None
        self.word_begin_label = word_begin_label
        self.word_end_label = word_end_label
        self.start_state = 0
        self.non_silence_state = 1

        if self.silence_probability:
            self.silence_state = 2
        else:
            self.silence_state = None

    def clear(self):
        self.pronunciations = []

    def to_int(self, word: str) -> int:
        """
        Look up a word in the word symbol table

        Parameters
        ----------
        word: str
            Word to look up

        Returns
        -------
        int
            Integer ID of word in symbol table
        """
        if self.word_table.member(word):
            return self.word_table.find(word)
        return self.word_table.find(self.oov_word)

    @property
    def specials_set(self) -> typing.Set[str]:
        """Special words, like the ``oov_word`` ``silence_word``, ``<s>``, and ``</s>``"""
        return {
            "#0",
            self.silence_word,
            "<s>",
            "</s>",
        }

    def load_pronunciations(self, file_name: typing.Union[pathlib.Path, str]) -> None:
        """
        Load pronunciations from a dictionary file and calculate necessary disambiguation symbols

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Path to lexicon file
        """
        non_silence_phones = set()
        words = set()
        oov_found = False
        for pron in parse_dictionary_file(file_name):
            if self.ignore_case:
                pron.orthography = pron.orthography.lower()
            if pron.orthography in self.specials_set:
                continue
            if pron.orthography == self.oov_word:
                oov_found = True
            phones = pron.pronunciation.split()
            non_silence_phones.update(phones)
            self.pronunciations.append(pron)
            words.add(pron.orthography)
        if not oov_found:
            self.pronunciations.append(
                Pronunciation(
                    self.oov_word,
                    self.oov_phone,
                    1.0,
                    None,
                    None,
                    None,
                    None,
                )
            )
        for s in sorted(words):
            self.word_table.add_symbol(s)
        for s in sorted(non_silence_phones):
            if self.position_dependent_phones:
                for pos in ["_S", "_B", "_E", "_I"]:
                    self.phone_table.add_symbol(s + pos)
            else:
                self.phone_table.add_symbol(s)

        self.compute_disambiguation_symbols()

    @property
    def disambiguation_symbols(self) -> typing.List[int]:
        """List of integer IDs for disambiguation symbols in the phone symbol table"""
        return [
            i
            for i in range(self.phone_table.num_symbols())
            if self.phone_table.find(i).startswith("#")
        ]

    @property
    def silence_symbols(self) -> typing.List[int]:
        """List of integer IDs for silence symbols in the phone symbol table"""
        return [
            i
            for i in range(self.phone_table.num_symbols())
            if self.phone_table.find(i) in {self.silence_phone, self.oov_phone}
        ]

    def compute_disambiguation_symbols(self):
        """Calculate the necessary disambiguation symbols for the lexicon"""
        subsequences = set()
        for pron in self.pronunciations:

            phones = pron.pronunciation.split()
            while len(phones) > 0:
                subsequences.add(" ".join(phones))
                phones = phones[:-1]
        last_used = collections.defaultdict(int)

        for pron in self.pronunciations:
            if pron.pronunciation in subsequences:
                last_used[pron.pronunciation] += 1
                pron.disambiguation = last_used[pron.pronunciation]

        self.max_disambiguation_symbol = max(
            self.max_disambiguation_symbol, max(last_used.values())
        )
        for x in range(self.max_disambiguation_symbol + 3):
            p = f"#{x}"
            self.phone_table.add_symbol(p)
        if self.disambiguation:
            self.silence_disambiguation_symbol = f"#{self.max_disambiguation_symbol + 1}"
        self.word_table.add_symbol("#0")
        self.word_table.add_symbol("<s>")
        self.word_table.add_symbol("</s>")

    @property
    def align_lexicon(self):
        if self._align_lexicon is None:
            lex = []
            word_symbol = self.to_int(self.silence_word)
            lex.append([word_symbol, word_symbol, self.phone_table.find(self.silence_phone)])
            for pron in self.pronunciations:
                word_symbol = self.to_int(pron.orthography)
                phones = pron.pronunciation.split()
                if self.position_dependent_phones:
                    if len(phones) == 1:
                        phones[0] += "_S"
                    else:
                        phones[0] += "_B"
                        for i in range(1, len(phones) - 1):
                            phones[i] += "_I"
                        phones[-1] += "_E"

                lex.append([word_symbol, word_symbol, *(self.phone_table.find(x) for x in phones)])

            self._align_lexicon = WordAlignLatticeLexiconInfo(lex)
        return self._align_lexicon

    def create_fsts(self, phonological_rule_fst: pynini.Fst = None):
        if self._fst is not None and self._align_fst is not None:
            return

        initial_silence_cost = 0
        initial_non_silence_cost = 0
        if self.initial_silence_probability:
            initial_silence_cost = -1 * math.log(self.initial_silence_probability)
            initial_non_silence_cost = -1 * math.log(1.0 - self.initial_silence_probability)

        final_silence_cost = 0
        final_non_silence_cost = 0
        if self.final_silence_correction:
            final_silence_cost = -math.log(self.final_silence_correction)
            final_non_silence_cost = -math.log(self.final_non_silence_correction)

        self.phone_table.find(self.silence_disambiguation_symbol)
        phone_eps_symbol = self.phone_table.find("<eps>")
        self.word_table.find(self.silence_word)
        self._fst = pynini.Fst()
        self._align_fst = pynini.Fst()
        # Start state = 0
        self._fst.add_state()
        self._align_fst.add_state()
        self._fst.set_start(self.start_state)

        # Non silence state = 1
        self._fst.add_state()  # Also loop state
        self._align_fst.add_state()

        # Silence state = 2
        if self.silence_probability:
            self._fst.add_state()
            self._align_fst.add_state()

        self._align_fst.set_start(self.start_state)
        # initial no silence
        self._fst.add_arc(
            self.start_state,
            pywrapfst.Arc(
                phone_eps_symbol,
                self.word_table.find(self.silence_word),
                pywrapfst.Weight(self._fst.weight_type(), initial_non_silence_cost),
                self.non_silence_state,
            ),
        )
        self._align_fst.add_arc(
            self.start_state,
            pywrapfst.Arc(
                phone_eps_symbol,
                self.word_table.find(self.silence_word),
                pywrapfst.Weight(self._align_fst.weight_type(), initial_non_silence_cost),
                self.non_silence_state,
            ),
        )
        # initial silence
        if self.silence_probability:
            self._fst.add_arc(
                self.start_state,
                pywrapfst.Arc(
                    self.phone_table.find(self.silence_phone),
                    self.word_table.find(self.silence_word),
                    pywrapfst.Weight(self._fst.weight_type(), initial_silence_cost),
                    self.silence_state,
                ),
            )
            self._align_fst.add_arc(
                self.start_state,
                pywrapfst.Arc(
                    self.phone_table.find(self.silence_phone),
                    self.word_table.find(self.silence_word),
                    pywrapfst.Weight(self._align_fst.weight_type(), initial_silence_cost),
                    self.silence_state,
                ),
            )

        for pron in self.pronunciations:
            self.add_pronunciation(pron, phonological_rule_fst)

        if self.silence_probability:
            if final_silence_cost > 0:
                self._fst.set_final(
                    self.silence_state,
                    pywrapfst.Weight(self._fst.weight_type(), final_silence_cost),
                )
                self._align_fst.set_final(
                    self.silence_state,
                    pywrapfst.Weight(self._align_fst.weight_type(), final_silence_cost),
                )
            else:
                self._fst.set_final(
                    self.silence_state, pywrapfst.Weight.one(self._fst.weight_type())
                )
                self._align_fst.set_final(
                    self.silence_state, pywrapfst.Weight.one(self._align_fst.weight_type())
                )
        if final_non_silence_cost > 0:
            self._fst.set_final(
                self.non_silence_state,
                pywrapfst.Weight(self._fst.weight_type(), final_non_silence_cost),
            )
            self._align_fst.set_final(
                self.non_silence_state,
                pywrapfst.Weight(self._align_fst.weight_type(), final_non_silence_cost),
            )
        else:
            self._fst.set_final(
                self.non_silence_state, pywrapfst.Weight.one(self._fst.weight_type())
            )
            self._align_fst.set_final(
                self.non_silence_state, pywrapfst.Weight.one(self._align_fst.weight_type())
            )

        if (
            self._fst.num_states()
            <= (self.silence_state if self.silence_probability else self.non_silence_state) + 1
            or self._fst.start() == pywrapfst.NO_STATE_ID
        ):
            num_words = self.word_table.num_symbols()
            num_phones = self.phone_table.num_symbols()
            num_pronunciations = len(self.pronunciations)
            raise LexiconError(
                f"There was an error compiling the lexicon "
                f"({num_words} words, {num_pronunciations} pronunciations, "
                f"{num_phones} phones)."
            )
        self._align_fst.arcsort("olabel")
        self._fst.arcsort("olabel")

    @property
    def base_silence_following_cost(self):
        base_silence_following_cost = 0
        if self.silence_probability:
            base_silence_following_cost = -math.log(self.silence_probability)
        return base_silence_following_cost

    @property
    def base_non_silence_following_cost(self):
        base_non_silence_following_cost = 0
        if self.silence_probability:
            base_non_silence_following_cost = -math.log(1 - self.silence_probability)
        return base_non_silence_following_cost

    @property
    def fst(self) -> pynini.Fst:
        """Compiled lexicon FST"""
        if self._fst is None:
            self.create_fsts()
        return self._fst

    def _create_word_fst(
        self, pronunciation: Pronunciation, phonological_rule_fst: pynini.Fst = None
    ):
        pron = pronunciation.pronunciation
        if self.position_dependent_phones:
            phones = pronunciation.pronunciation.split()
            if len(phones) == 1:
                phones[0] += "_S"
            else:
                phones[0] += "_B"
                phones[-1] += "_E"
                for i in range(1, len(phones) - 1):
                    phones[i] += "_I"
            pron = " ".join(phones)
        if not self.word_table.member(pronunciation.orthography):
            self.word_table.add_symbol(pronunciation.orthography)
        word_symbol = self.word_table.find(pronunciation.orthography)
        if self.disambiguation and pronunciation.disambiguation is not None:
            pron += f" #{pronunciation.disambiguation}"
        probability = pronunciation.probability
        weight = pywrapfst.Weight.one("tropical")
        if probability is not None:
            if probability < 0.01:
                probability = 0.01  # Dithering to ensure low probability entries
            weight = pywrapfst.Weight("tropical", abs(math.log(probability)))
        fst = pynini.accep(pron, weight=weight, token_type=self.phone_table)
        if phonological_rule_fst:
            fst = pynini.compose(phonological_rule_fst, fst)
        fst = pynini.arcmap(fst, map_type="output_epsilon")
        arcs = []
        for arc in fst.mutable_arcs(fst.start()):
            arc = arc.copy()
            arc.olabel = word_symbol
            arcs.append(arc)
        fst.delete_arcs(fst.start())
        for arc in arcs:
            fst.add_arc(fst.start(), arc)

        silence_before_cost = (
            -math.log(pronunciation.silence_before_correction)
            if pronunciation.silence_before_correction
            else 0.0
        )
        non_silence_before_cost = (
            -math.log(pronunciation.non_silence_before_correction)
            if pronunciation.non_silence_before_correction
            else 0.0
        )
        silence_following_cost = (
            -math.log(pronunciation.silence_after_probability)
            if pronunciation.silence_after_probability
            else self.base_silence_following_cost
        )
        non_silence_following_cost = (
            -math.log(1 - pronunciation.silence_after_probability)
            if pronunciation.silence_after_probability
            else self.base_non_silence_following_cost
        )
        initial_silence_fst = pynini.union(
            pynini.accep(
                self.silence_phone,
                weight=pywrapfst.Weight("tropical", silence_before_cost),
                token_type=self.phone_table,
            ),
            pynini.accep(
                "<eps>",
                weight=pywrapfst.Weight("tropical", non_silence_before_cost),
                token_type=self.phone_table,
            ),
        )

        initial_silence_fst = pynini.arcmap(initial_silence_fst, map_type="output_epsilon")
        final_silence_fst = pynini.union(
            pynini.accep(
                self.silence_phone,
                weight=pywrapfst.Weight("tropical", silence_following_cost),
                token_type=self.phone_table,
            ),
            pynini.accep(
                "<eps>",
                weight=pywrapfst.Weight("tropical", non_silence_following_cost),
                token_type=self.phone_table,
            ),
        )
        final_silence_fst = pynini.arcmap(final_silence_fst, map_type="output_epsilon")
        fst = initial_silence_fst + fst + final_silence_fst
        fst.optimize()
        return fst

    def add_pronunciation(
        self, pronunciation: Pronunciation, phonological_rule_fst: pynini.Fst = None
    ):
        if (pronunciation.orthography, pronunciation.pronunciation) in self._cached_pronunciations:
            return
        phones = pronunciation.pronunciation.split()
        if self.position_dependent_phones:
            if len(phones) == 1:
                phones[0] += "_S"
            else:
                phones[0] += "_B"
                phones[-1] += "_E"
                for i in range(1, len(phones) - 1):
                    phones[i] += "_I"
        new_phones = ", ".join(sorted({x for x in phones if not self.phone_table.member(x)}))
        if new_phones:
            raise Exception(
                f"The pronunciation '{pronunciation}' had the following phones not in the symbol table: {new_phones}"
            )
        pron = " ".join(phones)
        fst = pynini.accep(pron, token_type=self.phone_table)
        if phonological_rule_fst:
            fst = pynini.compose(phonological_rule_fst, fst)
        fst.rmepsilon()
        self._cached_pronunciations.add((pronunciation.orthography, pronunciation.pronunciation))
        if not self.word_table.member(pronunciation.orthography):
            self.word_table.add_symbol(pronunciation.orthography)
        word_symbol = self.word_table.find(pronunciation.orthography)
        word_eps_symbol = self.word_table.find("<eps>")
        phone_eps_symbol = self.phone_table.find("<eps>")
        silence_before_cost = (
            -math.log(pronunciation.silence_before_correction)
            if pronunciation.silence_before_correction
            else 0.0
        )
        non_silence_before_cost = (
            -math.log(pronunciation.non_silence_before_correction)
            if pronunciation.non_silence_before_correction
            else 0.0
        )
        silence_following_cost = (
            -math.log(pronunciation.silence_after_probability)
            if pronunciation.silence_after_probability
            else self.base_silence_following_cost
        )
        non_silence_following_cost = (
            -math.log(1 - pronunciation.silence_after_probability)
            if pronunciation.silence_after_probability
            else self.base_non_silence_following_cost
        )
        probability = pronunciation.probability
        if probability is None:
            probability = 1
        elif probability < 0.01:
            probability = 0.01  # Dithering to ensure low probability entries
        pron_cost = abs(math.log(probability))
        start_index = self.fst.num_states() - 1
        align_start_index = self.align_fst.num_states()
        num_new_states = fst.num_states() - 1
        self._fst.add_states(num_new_states)
        self._align_fst.add_states(num_new_states + 2)

        # FST arcs
        for state in fst.states():
            for arc in fst.arcs(state):
                if state == fst.start():
                    # No silence before the pronunciation
                    self._fst.add_arc(
                        self.non_silence_state,
                        pywrapfst.Arc(
                            arc.ilabel,
                            word_symbol,
                            pywrapfst.Weight(
                                self._fst.weight_type(), pron_cost + non_silence_before_cost
                            ),
                            arc.nextstate + start_index,
                        ),
                    )
                    # Silence before the pronunciation
                    if self.silence_probability:
                        self._fst.add_arc(
                            self.silence_state,
                            pywrapfst.Arc(
                                arc.ilabel,
                                word_symbol,
                                pywrapfst.Weight(
                                    self._fst.weight_type(), pron_cost + silence_before_cost
                                ),
                                arc.nextstate + start_index,
                            ),
                        )

                    # No silence before the pronunciation
                    self._align_fst.add_arc(
                        self.non_silence_state,
                        pywrapfst.Arc(
                            self.phone_table.find(self.word_begin_label),
                            word_symbol,
                            pywrapfst.Weight(
                                self._fst.weight_type(), pron_cost + non_silence_before_cost
                            ),
                            arc.nextstate + align_start_index - 1,
                        ),
                    )
                    # Silence before the pronunciation
                    if self.silence_probability:
                        self._align_fst.add_arc(
                            self.silence_state,
                            pywrapfst.Arc(
                                self.phone_table.find(self.word_begin_label),
                                word_symbol,
                                pywrapfst.Weight(
                                    self._fst.weight_type(), pron_cost + silence_before_cost
                                ),
                                arc.nextstate + align_start_index - 1,
                            ),
                        )
                else:
                    self._fst.add_arc(
                        state + start_index,
                        pywrapfst.Arc(
                            arc.ilabel,
                            word_eps_symbol,
                            arc.weight,
                            arc.nextstate + start_index,
                        ),
                    )
                self._align_fst.add_arc(
                    state + align_start_index,
                    pywrapfst.Arc(
                        arc.ilabel,
                        word_eps_symbol,
                        arc.weight,
                        arc.nextstate + align_start_index,
                    ),
                )

        if self.disambiguation and pronunciation.disambiguation is not None:
            self._fst.add_state()
            self._fst.add_arc(
                num_new_states + start_index,
                pywrapfst.Arc(
                    self.phone_table.find(f"#{pronunciation.disambiguation}"),
                    word_eps_symbol,
                    pywrapfst.Weight(self._fst.weight_type(), non_silence_following_cost),
                    num_new_states + start_index + 1,
                ),
            )
            start_index += 1

        # No silence following the pronunciation
        self._fst.add_arc(
            num_new_states + start_index,
            pywrapfst.Arc(
                self.phone_table.find(self.silence_disambiguation_symbol),
                word_eps_symbol,
                pywrapfst.Weight(self._fst.weight_type(), non_silence_following_cost),
                self.non_silence_state,
            ),
        )
        # Silence following the pronunciation
        if self.silence_probability:
            self._fst.add_arc(
                num_new_states + start_index,
                pywrapfst.Arc(
                    self.phone_table.find(self.silence_phone),
                    word_eps_symbol,
                    pywrapfst.Weight(self._fst.weight_type(), silence_following_cost),
                    self.silence_state,
                ),
            )
        self._align_fst.add_arc(
            num_new_states + align_start_index,
            pywrapfst.Arc(
                self.phone_table.find(self.word_end_label),
                word_eps_symbol,
                pywrapfst.Weight.one(self._align_fst.weight_type()),
                num_new_states + align_start_index + 1,
            ),
        )

        # No silence following the pronunciation
        self._align_fst.add_arc(
            num_new_states + align_start_index + 1,
            pywrapfst.Arc(
                phone_eps_symbol,
                word_eps_symbol,
                pywrapfst.Weight(self._fst.weight_type(), non_silence_following_cost),
                self.non_silence_state,
            ),
        )
        # Silence following the pronunciation
        if self.silence_probability:
            self._align_fst.add_arc(
                num_new_states + align_start_index + 1,
                pywrapfst.Arc(
                    self.phone_table.find(self.silence_phone),
                    word_eps_symbol,
                    pywrapfst.Weight(self._fst.weight_type(), silence_following_cost),
                    self.silence_state,
                ),
            )

    @property
    def kaldi_fst(self) -> VectorFst:
        return VectorFst.from_pynini(self.fst)

    def load_l_from_file(
        self,
        l_fst_path: typing.Union[pathlib.Path, str],
    ) -> None:
        """
        Read g.fst from file

        Parameters
        ----------
        l_fst_path: :class:`~pathlib.Path` or str
            Path to read HCLG.fst
        """
        self._fst = pynini.Fst.read(str(l_fst_path))

    def load_l_align_from_file(
        self,
        l_fst_path: typing.Union[pathlib.Path, str],
    ) -> None:
        """
        Read g.fst from file

        Parameters
        ----------
        l_fst_path: :class:`~pathlib.Path` or str
            Path to read HCLG.fst
        """
        self._align_fst = pynini.Fst.read(str(l_fst_path))

    @property
    def align_fst(self) -> pynini.Fst:
        """Compiled FST for aligning lattices when `position_dependent_phones` is False"""
        if self._align_fst is None:
            self.create_fsts()
        return self._align_fst

    def _create_pronunciation_string(
        self,
        word_symbols: typing.List[int],
        phone_symbols: typing.List[int],
        transcription: bool = False,
    ):
        word_begin_symbol = self.phone_table.find(self.word_begin_label)
        word_end_symbol = self.phone_table.find(self.word_end_label)
        text = " ".join(self.word_table.find(x) for x in word_symbols)
        if transcription:
            if len(word_symbols) > 1:
                text, final_word = text.rsplit(maxsplit=1)
            else:
                final_word = text
        phone_to_word = pynini.compose(
            self.align_fst, pynini.accep(text, token_type=self.word_table)
        )
        if transcription:
            final_word_phone_to_word = pynini.compose(
                self.align_fst, pynini.accep(final_word, token_type=self.word_table)
            )
            infinity_weight = pywrapfst.Weight(final_word_phone_to_word.weight_type(), "infinity")
            final_word_phone_to_word = pynini.determinize(final_word_phone_to_word)
            final_states = []
            for i in range(final_word_phone_to_word.num_states()):
                if final_word_phone_to_word.final(i) != infinity_weight:
                    final_states.append(i)
                else:
                    final_word_phone_to_word.set_final(
                        i, pywrapfst.Weight.one(final_word_phone_to_word.weight_type())
                    )
            extra_state = final_word_phone_to_word.add_state()
            final_word_phone_to_word.set_final(
                extra_state, pywrapfst.Weight.one(final_word_phone_to_word.weight_type())
            )
            for state in final_states:
                final_word_phone_to_word.add_arc(
                    state,
                    pywrapfst.Arc(
                        word_begin_symbol,
                        self.phone_table.find("<eps>"),
                        pywrapfst.Weight(final_word_phone_to_word.weight_type(), 10),
                        extra_state,
                    ),
                )
            for i in range(self.phone_table.num_symbols()):
                if self.phone_table.find(i) == "<eps>":
                    continue
                if self.phone_table.find(i).startswith(self.silence_phone):
                    continue
                if self.phone_table.find(i).startswith("#"):
                    continue

                final_word_phone_to_word.add_arc(
                    extra_state,
                    pywrapfst.Arc(
                        i,
                        self.phone_table.find("<eps>"),
                        pywrapfst.Weight(final_word_phone_to_word.weight_type(), 10),
                        extra_state,
                    ),
                )
            if len(word_symbols) > 1:
                phone_to_word = pynini.concat(phone_to_word, final_word_phone_to_word)
            else:
                phone_to_word = final_word_phone_to_word

        phone_fst = pynini.Fst()
        current_state = phone_fst.add_state()
        phone_fst.set_start(current_state)
        for symbol in phone_symbols:
            next_state = phone_fst.add_state()
            phone_fst.add_arc(
                current_state,
                pywrapfst.Arc(
                    symbol, symbol, pywrapfst.Weight.one(phone_fst.weight_type()), next_state
                ),
            )
            current_state = next_state
        phone_fst.set_final(current_state, pywrapfst.Weight.one(phone_fst.weight_type()))

        for s in range(current_state + 1):
            phone_fst.add_arc(
                s,
                pywrapfst.Arc(
                    word_end_symbol,
                    word_end_symbol,
                    pywrapfst.Weight.one(phone_fst.weight_type()),
                    s,
                ),
            )
            phone_fst.add_arc(
                s,
                pywrapfst.Arc(
                    word_begin_symbol,
                    word_begin_symbol,
                    pywrapfst.Weight.one(phone_fst.weight_type()),
                    s,
                ),
            )
        phone_fst.arcsort("olabel")
        if transcription:
            inf_weight = pywrapfst.Weight(phone_fst.weight_type(), "infinity")
            for state in range(phone_fst.num_states()):
                if phone_fst.final(state) != inf_weight:
                    phone_fst.set_final(state, pywrapfst.Weight(phone_fst.weight_type(), 100))

        lattice = pynini.compose(phone_fst, phone_to_word)

        projection = pynini.shortestpath(lattice).project("input")
        if projection.start() == pywrapfst.NO_SYMBOL:
            phone_fst.set_input_symbols(self.phone_table)
            phone_fst.set_output_symbols(self.phone_table)
            phone_to_word.set_input_symbols(self.phone_table)
            phone_to_word.set_output_symbols(self.word_table)
            raise PhonesToPronunciationsError(
                text,
                " ".join(self.phone_table.find(x) for x in phone_symbols),
                phone_fst,
                phone_to_word,
            )
        path_string = projection.string(self.phone_table)
        if self.position_dependent_phones:
            path_string = re.sub(r"_[SIBE]\b", "", path_string)

        path_string = re.sub(f" {self.word_end_label}$", "", path_string)
        path_string = path_string.replace(
            f"{self.word_end_label} {self.word_end_label}", self.word_end_label
        )
        path_string = path_string.replace(
            f"{self.word_end_label} {self.word_begin_label}", self.word_begin_label
        )
        path_string = path_string.replace(f"{self.word_end_label}", self.word_begin_label)
        path_string = re.sub(f"^{self.word_begin_label} ", "", path_string)
        word_splits = [x for x in re.split(rf" ?{self.word_begin_label} ?", path_string)]
        return word_splits

    def phones_to_pronunciations(
        self,
        word_symbols: typing.List[int],
        intervals: typing.List[CtmInterval],
        transcription: bool = False,
        text: str = None,
    ) -> HierarchicalCtm:
        """
        Combine sequences of word symbols and phone intervals in a :class:`~kalpy.gmm.data.HierarchicalCtm`

        Parameters
        ----------
        word_symbols: list[int]
            Sequence of word symbols
        intervals: list[:class:`~_kalpy.util.CtmInterval`]
            Sequence of phone intervals
        transcription: bool
            Flag for whether the sequences were generated from a transcription lattice
        text: str, optional
            Original utterance text to map word symbols to

        Returns
        -------
        :class:`~kalpy.gmm.data.HierarchicalCtm`
            Hierarchical CTM object with word and phone intervals for the utterance
        """

        phones = [x.symbol for x in intervals]
        word_splits = self._create_pronunciation_string(
            word_symbols,
            phones,
            transcription=transcription,
        )

        actual_words = [self.word_table.find(x) for x in word_symbols]
        if not text:
            text = " ".join(actual_words)
        word_intervals = []
        current_phone_index = 0
        current_word_index = 0
        for i, w in enumerate(actual_words):
            if current_word_index >= len(word_splits):
                break
            pron = word_splits[current_word_index]
            word_symbol = word_symbols[i]
            if pron == self.silence_phone:
                word_intervals.append(
                    WordCtmInterval(
                        self.silence_word,
                        self.word_table.find(self.silence_word),
                        intervals[current_phone_index : current_phone_index + 1],
                    )
                )
                current_word_index += 1
                current_phone_index += 1
                if current_word_index >= len(word_splits):
                    break
                pron = word_splits[current_word_index]
            if pron:
                phones = pron.split()
                word_intervals.append(
                    WordCtmInterval(
                        w,
                        word_symbol,
                        intervals[current_phone_index : current_phone_index + len(phones)],
                    )
                )
                current_phone_index += len(phones)
            current_word_index += 1
        if current_word_index < len(word_splits):
            pron = word_splits[current_word_index]
            if pron == self.silence_phone:
                word_intervals.append(
                    WordCtmInterval(
                        self.silence_word,
                        self.word_table.find(self.silence_word),
                        intervals[current_phone_index : current_phone_index + 1],
                    )
                )
        if not word_intervals[-1].phones:
            del word_intervals[-1]
        return HierarchicalCtm(word_intervals, text=text)


class G2PCompiler(LexiconCompiler):
    use_g2p = True

    def __init__(
        self,
        fst: pynini.Fst,
        grapheme_table: pywrapfst.SymbolTable,
        phone_table: pywrapfst.SymbolTable,
        silence_phone: str = "sil",
        silence_word: str = "<eps>",
        align_fst: typing.Optional[pynini.Fst] = None,
        position_dependent_phones: bool = False,
    ):
        self._fst = fst
        self._align_fst = align_fst
        self._align_fst.invert()
        self.word_table = grapheme_table
        self.phone_table = phone_table
        self.silence_phone = silence_phone
        self.silence_word = silence_word
        self.word_begin_label = "#1"
        self.word_end_label = "#2"
        self.position_dependent_phones = position_dependent_phones
        self.disambiguation = False

    def phones_to_pronunciations(
        self,
        word_symbols: typing.List[int],
        intervals: typing.List[CtmInterval],
        transcription: bool = False,
        text: str = None,
    ) -> HierarchicalCtm:
        phone_symbols = [x.symbol for x in intervals]
        word_symbols = [self.word_table.find(x) for x in text.split()]
        word_splits = self._create_pronunciation_string(
            word_symbols,
            phone_symbols,
            transcription=transcription,
        )

        # Might need some better logic
        actual_words = [x.replace(" ", "") for x in text.split("<space>")]
        word_intervals = []
        current_phone_index = 0
        current_word_index = 0
        for w in actual_words:
            pron = word_splits[current_word_index]
            if pron == self.silence_phone:
                word_intervals.append(
                    WordCtmInterval(
                        self.silence_word,
                        0,
                        intervals[current_phone_index : current_phone_index + 1],
                    )
                )
                current_word_index += 1
                current_phone_index += 1
                pron = word_splits[current_word_index]

            phones = pron.split()
            word_intervals.append(
                WordCtmInterval(
                    w, 0, intervals[current_phone_index : current_phone_index + len(phones)]
                )
            )
            current_phone_index += len(phones)
            current_word_index += 1
        if current_word_index != len(word_splits):
            pron = word_splits[current_word_index]
            if pron == self.silence_phone:
                word_intervals.append(
                    WordCtmInterval(
                        self.silence_word,
                        0,
                        intervals[current_phone_index : current_phone_index + 1],
                    )
                )
        return HierarchicalCtm(word_intervals, text=text)



"""Classes for training GMM models"""
from __future__ import annotations

import logging
import pathlib
import typing

from _kalpy import gmm, hmm, tree
from _kalpy.matrix import DoubleVector
from _kalpy.util import Output
from kalpy.feat.data import FeatureArchive
from kalpy.gmm.data import AlignmentArchive
from kalpy.gmm.utils import read_gmm_model

logger = logging.getLogger("kalpy.train")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class GmmStatsAccumulator:
    def __init__(self, acoustic_model_path: typing.Union[pathlib.Path, str]):
        self.acoustic_model_path = str(acoustic_model_path)
        self.transition_accs = DoubleVector()
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.gmm_accs = gmm.AccumAmDiagGmm()
        self.transition_model.InitStats(self.transition_accs)
        self.gmm_accs.Init(self.acoustic_model, gmm.kGmmAll)
        self.callback_frequency = 100

    def accumulate_stats(
        self,
        feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
        callback: typing.Callable = None,
        ignored_keys: typing.Collection[str] = None,
    ):
        tot_like = 0.0
        tot_t = 0.0
        num_done = 0
        for utterance_id, feats in feature_archive:
            if ignored_keys is not None and utterance_id in ignored_keys:
                logger.warning(f"Ignoring {utterance_id}")
                continue
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utterance_id} due to zero-length features")
                continue
            try:
                alignment = alignment_archive[utterance_id]
            except KeyError:
                logger.warning(f"Skipping {utterance_id} due to missing alignment")
                continue
            tot_like_this_file = self.gmm_accs.acc_stats(
                self.acoustic_model, self.transition_model, alignment.alignment, feats
            )
            self.transition_model.acc_stats(alignment.alignment, self.transition_accs)
            num_done += 1
            tot_like += tot_like_this_file
            tot_t += len(alignment.alignment)
            if num_done % self.callback_frequency == 0:
                if callback:
                    callback(self.callback_frequency)
                logger.info(
                    f"Processed {num_done} utterances; for utterance "
                    f"{alignment.utterance_id} avg. like is "
                    f"{tot_like_this_file/len(alignment.alignment)} "
                    f"over {len(alignment.alignment)} frames."
                )
        if callback is not None and num_done % self.callback_frequency:
            callback(num_done % self.callback_frequency)
        logger.info(f"Done {num_done} files.")
        if tot_t:
            logger.info(
                f"Overall avg like per frame (Gaussian only) = {tot_like/tot_t} over {tot_t} frames."
            )

    def export_stats(
        self,
        file_name: typing.Union[pathlib.Path, str],
        feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
    ):
        file_name = str(file_name)
        self.accumulate_stats(feature_archive, alignment_archive)
        ko = Output(file_name, True)
        self.transition_accs.Write(ko.Steam(), True)
        self.gmm_accs.Write(ko.Steam(), True)
        ko.Close()
        logger.info("Written accs.")


class TreeStatsAccumulator:
    def __init__(
        self,
        acoustic_model_path: typing.Union[pathlib.Path, str],
        var_floor: float = 0.01,
        context_width: int = 3,
        central_position: int = 1,
        context_independent_symbols: typing.List[int] = None,
        phone_map: typing.List[int] = None,
    ):
        self.acoustic_model_path = str(acoustic_model_path)
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.tree_stats_opts = hmm.AccumulateTreeStatsOptions()
        self.tree_stats_opts.var_floor = var_floor
        self.tree_stats_opts.context_width = context_width
        self.tree_stats_opts.central_position = central_position
        self.tree_stats_info = hmm.AccumulateTreeStatsInfo(self.tree_stats_opts)
        self.tree_stats_info.ci_phones = context_independent_symbols
        if phone_map:
            self.tree_stats_info.ci_phones = phone_map
        self.tree_stats = {}
        self.callback_frequency = 100

    def accumulate_stats(
        self,
        feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
        callback: typing.Callable = None,
    ):
        num_done = 0
        for utterance_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utterance_id} due to zero-length features")
                continue
            try:
                alignment = alignment_archive[utterance_id]
            except KeyError:
                logger.warning(f"Skipping {utterance_id} due to missing alignment")
                continue
            if callback is not None and num_done % self.callback_frequency == 0:
                callback(self.callback_frequency)
            stats = hmm.accumulate_tree_stats(
                self.transition_model, self.tree_stats_info, alignment.alignment, feats
            )
            for e, c in stats:
                e = tuple(e)
                if e not in self.tree_stats:
                    self.tree_stats[e] = c
                else:
                    self.tree_stats[e].Add(c)
            num_done += 1
        if callback is not None and num_done % self.callback_frequency:
            callback(num_done % self.callback_frequency)
        logger.info(f"Done {num_done} files.")

    def export_stats(
        self,
        file_name: typing.Union[pathlib.Path, str],
        feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
    ):
        file_name = str(file_name)
        self.accumulate_stats(feature_archive, alignment_archive)
        ko = Output(file_name, True)
        tree.WriteBuildTreeStats(ko.Steam(), True, [x for x in self.tree_stats.items()])
        ko.Close()
        logger.info("Written tree stats.")


class TwoFeatsStatsAccumulator:
    def __init__(self, acoustic_model_path: typing.Union[pathlib.Path, str]):
        self.acoustic_model_path = str(acoustic_model_path)
        self.transition_accs = DoubleVector()
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.gmm_accs = gmm.AccumAmDiagGmm()
        self.transition_model.InitStats(self.transition_accs)
        self.gmm_accs.Init(self.acoustic_model, gmm.kGmmAll)
        self.callback_frequency = 100

    def accumulate_stats(
        self,
        first_feature_archive: FeatureArchive,
        second_feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
        callback: typing.Callable = None,
    ):
        num_done = 0
        tot_like = 0.0
        for (utterance_id, first_feats), (second_utterance_id, second_feats) in zip(
            first_feature_archive, second_feature_archive
        ):
            assert utterance_id == second_utterance_id
            if first_feats.NumRows() == 0:
                logger.warning(
                    f"Skipping {utterance_id} due to zero-length features in first archive"
                )
                continue
            if second_feats.NumRows() == 0:
                logger.warning(
                    f"Skipping {utterance_id} due to zero-length features in second archive"
                )
                continue
            try:
                alignment = alignment_archive[utterance_id]
            except KeyError:
                logger.warning(f"Skipping {utterance_id} due to missing alignment")
                continue
            if callback is not None and num_done % self.callback_frequency == 0:
                callback(self.callback_frequency)
            post = hmm.AlignmentToPosterior(alignment.alignment)
            pdf_post = hmm.convert_posterior_to_pdfs(self.transition_model, post)
            tot_like_this_file = self.gmm_accs.acc_twofeats(
                self.acoustic_model, pdf_post, first_feats, second_feats
            )
            self.transition_model.acc_twofeats(
                post, first_feats, second_feats, self.transition_accs
            )
            num_done += 1
            tot_like += tot_like_this_file
        if callback is not None and num_done % self.callback_frequency:
            callback(num_done % self.callback_frequency)
        logger.info(f"Done {num_done} files.")

    def export_stats(
        self,
        file_name: str,
        first_feature_archive: FeatureArchive,
        second_feature_archive: FeatureArchive,
        alignment_archive: AlignmentArchive,
    ):
        file_name = str(file_name)
        self.accumulate_stats(first_feature_archive, second_feature_archive, alignment_archive)
        ko = Output(file_name, True)
        self.transition_accs.Write(ko.Steam(), True)
        self.gmm_accs.Write(ko.Steam(), True)
        ko.Close()
        logger.info("Written accs.")



from __future__ import annotations

import os.path
import pathlib
import typing

from _kalpy.gmm import AmDiagGmm
from _kalpy.hmm import HmmTopology, TransitionModel
from _kalpy.tree import ContextDependency
from _kalpy.util import Input, Output, ReadKaldiObject
from kalpy.exceptions import ReadError


def read_transition_model(model_path: typing.Union[str, pathlib.Path]) -> TransitionModel:
    if not os.path.exists(model_path):
        raise ReadError(f"The specified model path {model_path} does not exist.")
    ki = Input()
    ki.Open(str(model_path), True)
    transition_model = TransitionModel()
    transition_model.Read(ki.Stream(), True)
    ki.Close()
    return transition_model


def read_gmm_model(
    model_path: typing.Union[str, pathlib.Path]
) -> typing.Tuple[TransitionModel, AmDiagGmm]:
    if not os.path.exists(model_path):
        raise ReadError(f"The specified model path {model_path} does not exist.")
    ki = Input()
    ki.Open(str(model_path), True)
    transition_model = TransitionModel()
    transition_model.Read(ki.Stream(), True)
    acoustic_model = AmDiagGmm()
    acoustic_model.Read(ki.Stream(), True)
    ki.Close()
    return transition_model, acoustic_model


def read_topology(topo_path: typing.Union[str, pathlib.Path]) -> HmmTopology:
    if not os.path.exists(topo_path):
        raise ReadError(f"The specified topo path {topo_path} does not exist.")
    ki = Input()
    ki.Open(str(topo_path), False)
    topo = HmmTopology()
    topo.Read(ki.Stream(), False)
    ki.Close()
    return topo


def write_gmm_model(
    model_path: typing.Union[str, pathlib.Path],
    transition_model: TransitionModel,
    acoustic_model: AmDiagGmm,
    binary: bool = True,
) -> None:
    ko = Output(str(model_path), binary)
    transition_model.Write(ko.Stream(), binary)
    acoustic_model.Write(ko.Stream(), binary)
    ko.Close()


def write_tree(
    tree_path: typing.Union[str, pathlib.Path], tree: ContextDependency, binary: bool = True
) -> None:
    ko = Output(str(tree_path), binary)
    tree.Write(ko.Stream(), binary)
    ko.Close()


def read_tree(tree_path: typing.Union[str, pathlib.Path]) -> ContextDependency:
    if not os.path.exists(tree_path):
        raise ReadError(f"The specified tree path {tree_path} does not exist.")
    tree = ContextDependency()
    ReadKaldiObject(str(tree_path), tree)
    return tree



"""Classes for GMM alignment"""
from __future__ import annotations

import logging
import pathlib
import sys
import traceback
import typing

from _kalpy.fstext import VectorFst
from _kalpy.gmm import gmm_align_compiled, gmm_align_reference_phones
from _kalpy.matrix import FloatMatrix
from _kalpy.util import (
    BaseFloatVectorWriter,
    Int32VectorWriter,
    RandomAccessInt32VectorVectorReader,
)
from kalpy.decoder.data import FstArchive
from kalpy.feat.data import FeatureArchive
from kalpy.gmm.data import Alignment
from kalpy.gmm.utils import read_gmm_model
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.align")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class GmmAligner:
    """
    Wrapper class around aligning with a GMM acoustic model

    Parameters
    ----------
    acoustic_model_path: str or :class:`~pathlib.Path`
        Path to acoustic model
    beam: int
        Size of the beam to use in decoding, defaults to 10
    retry_beam : int
        Size of the beam to use in decoding if it fails with the initial beam width, defaults to 40
    transition_scale : float
        Transition scale, defaults to 1.0
    acoustic_scale : float
        Acoustic scale, defaults to 0.1
    self_loop_scale : float
        Self-loop scale, defaults to 0.1
    careful : bool
        Flag for extra error checking on reaching final state, defaults to False
    disambiguation_symbols : list[int], optional
        List of symbols in phone table that correspond to disambiguation symbols
    """

    def __init__(
        self,
        acoustic_model_path: typing.Union[pathlib.Path, str],
        acoustic_scale: float = 1.0,
        transition_scale: float = 1.0,
        self_loop_scale: float = 1.0,
        beam: int = 10,
        retry_beam: int = 40,
        careful: bool = False,
        disambiguation_symbols: typing.List[int] = None,
        silence_phones: typing.List[int] = None,
    ):
        self.acoustic_model_path = str(acoustic_model_path)
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.acoustic_scale = acoustic_scale
        self.transition_scale = transition_scale
        self.self_loop_scale = self_loop_scale
        self.beam = beam
        self.retry_beam = retry_beam
        self.careful = careful
        if silence_phones is None:
            silence_phones = set()
        self.silence_phones = silence_phones

        self.num_done = 0
        self.num_error = 0
        self.num_retry = 0
        self.total_likelihood = 0
        self.total_frames = 0
        self.disambiguation_symbols = (
            disambiguation_symbols if disambiguation_symbols is not None else []
        )
        if self.beam >= self.retry_beam:
            self.retry_beam = 4 * self.beam

    def boost_silence(self, silence_weight: float, only_silence: bool = True):
        """
        Function to boost probabilities associated with silence states in the GMM

        Parameters
        ----------
        silence_weight: float
            Factor to boost silence by, 1.0 means no boosting, less than 1.0 decreases silence state probability,
            greater than 1.0 increases silence probability
        silence_phones: list[int]
            Phone IDs corresponding to silence phones
        only_silence: bool
            Flag for whether to boost only states that are associated with silence phones and not speech phones
        """
        if only_silence:
            self.acoustic_model.boost_only_silence(
                self.transition_model, self.silence_phones, silence_weight
            )
        else:
            self.acoustic_model.boost_silence(
                self.transition_model, self.silence_phones, silence_weight
            )

    def align_utterance(
        self,
        training_graph: VectorFst,
        features: FloatMatrix,
        utterance_id: str = None,
        reference_phones: typing.List[typing.List[int]] = None,
        beam: int = None,
        retry_beam: int = None,
        boost_silence: float = None,
        acoustic_scale: float = None,
        transition_scale: float = None,
        self_loop_scale: float = None,
    ) -> typing.Optional[Alignment]:
        """

        Parameters
        ----------
        training_graph: :class:`~_kalpy.fstext.VectorFst`
            Training graph for the utterance
        features: :class:`~_kalpy.matrix.FloatMatrix`
            Feature matrix for the utterance
        utterance_id: str, optional
            Identifier for the utterance to store
        reference_phones: list[list[int]], optional
            Sequence of phones to use as reference and constrain alignments to only transition IDs corresponding to the
            phones specified for each frame

        Returns
        -------
        :class:`~kalpy.gmm.data.Alignment` or None
            Alignment object with list of transition IDs and other information for the utterance. If the utterance could
            not be aligned, returns None

        """
        if boost_silence is not None:
            self.boost_silence(boost_silence)
        alignment_parameters = {
            "beam": self.beam if not beam else beam,
            "retry_beam": self.retry_beam if not retry_beam else retry_beam,
            "acoustic_scale": self.acoustic_scale if not acoustic_scale else acoustic_scale,
            "transition_scale": self.transition_scale
            if not transition_scale
            else transition_scale,
            "self_loop_scale": self.self_loop_scale if not self_loop_scale else self_loop_scale,
        }
        if alignment_parameters["retry_beam"] <= alignment_parameters["beam"]:
            alignment_parameters["retry_beam"] = alignment_parameters["beam"] * 4
        if reference_phones is None:
            (
                alignment,
                words,
                likelihood,
                per_frame_log_likelihoods,
                successful,
                retried,
            ) = gmm_align_compiled(
                self.transition_model,
                self.acoustic_model,
                training_graph,
                features,
                careful=self.careful,
                **alignment_parameters,
            )
        else:
            logger.debug(f"Using reference phones in alignment for {utterance_id}")
            assert len(reference_phones) == features.NumRows()
            (
                alignment,
                words,
                likelihood,
                per_frame_log_likelihoods,
                successful,
                retried,
            ) = gmm_align_reference_phones(
                self.transition_model,
                self.acoustic_model,
                training_graph,
                features,
                reference_phones,
                careful=self.careful,
                **alignment_parameters,
            )
            if successful:
                redo = False
                alignment_phones = [
                    self.transition_model.TransitionIdToPhone(x) for x in alignment
                ]
                if len(reference_phones) != len(alignment):
                    logger.debug(
                        f"Mismatched alignment and reference length: {len(alignment)} vs {len(reference_phones)}"
                    )
                for i in range(len(reference_phones)):
                    if (
                        alignment_phones[i] not in reference_phones[i]
                        and -1 not in reference_phones[i]
                    ):
                        logger.debug(
                            f"Mismatch in frame {i}! {alignment_phones[i]} should be {reference_phones[i]}"
                        )
                        redo = True
                        reference_phones[i].append(-1)
                if redo:
                    logger.debug(f"Redoing {utterance_id} with more lenient reference phones")
                    (
                        alignment,
                        words,
                        likelihood,
                        per_frame_log_likelihoods,
                        successful,
                        retried,
                    ) = gmm_align_reference_phones(
                        self.transition_model,
                        self.acoustic_model,
                        training_graph,
                        features,
                        reference_phones,
                        careful=self.careful,
                        **alignment_parameters,
                    )
                    for i in range(len(reference_phones)):
                        if (
                            alignment_phones[i] not in reference_phones[i]
                            and -1 not in reference_phones[i]
                        ):
                            logger.debug(
                                f"Mismatch in frame {i}! {alignment_phones[i]} should be {reference_phones[i]}"
                            )
        if boost_silence is not None:
            self.boost_silence(1 / boost_silence)
        if not successful:
            return None
        if retried and utterance_id:
            logger.debug(f"Retried {utterance_id}")
        return Alignment(utterance_id, alignment, words, likelihood, per_frame_log_likelihoods)

    def align_utterances(
        self,
        training_graph_archive: FstArchive,
        feature_archive: FeatureArchive,
        reference_phone_archive: RandomAccessInt32VectorVectorReader = None,
        utterance_parameters: typing.Dict[str, typing.Dict[str, typing.Any]] = None,
    ) -> typing.Generator[Alignment]:
        """
        Function for aligning all utterances in a training graph archive and feature archive

        Parameters
        ----------
        training_graph_archive: :class:`~kalpy.decoder.data.FstArchive`
            Archive of training graph FSTs per utterance
        feature_archive: :class:`~kalpy.feat.data.FeatureArchive`
            Archive of feature matrices per utterance
        reference_phone_archive: :class:`~_kalpy.util.RandomAccessInt32VectorVectorReader`, optional
            Archive of reference phone sequences per utterance

        Yields
        ------
        :class:`~kalpy.gmm.data.Alignment`
            Successful alignments, skipping utterances that could not be aligned
        """
        logger.debug(f"Aligning with {self.acoustic_model_path}")
        num_done = 0
        num_error = 0
        total_frames = 0
        total_likelihood = 0
        if utterance_parameters is None:
            utterance_parameters = {}
        for utterance_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utterance_id} due to zero-length features")
                continue
            try:
                training_graph = training_graph_archive[utterance_id]
            except KeyError:
                logger.warning(f"Skipping {utterance_id} due to missing training graph")
                continue

            reference_phones = None
            if reference_phone_archive is not None:
                if reference_phone_archive.HasKey(utterance_id):
                    reference_phones = reference_phone_archive.Value(utterance_id)
            try:
                logger.debug(f"Processing {utterance_id}")
                parameters = utterance_parameters.get(utterance_id, {})
                alignment = self.align_utterance(
                    training_graph, feats, utterance_id, reference_phones, **parameters
                )
                if alignment is None:
                    yield utterance_id, None
                    num_error += 1
                    continue
                yield alignment
                total_likelihood += alignment.likelihood
                total_frames += len(alignment.alignment)
                num_done += 1
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                logger.warning(f"Error on {utterance_id}: {e}")
                traceback_lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
                logger.debug("\n".join(traceback_lines))
        if total_frames:
            logger.info(
                f"Overall log-likelihood per frame is {total_likelihood / total_frames} over {total_frames} frames."
            )
        logger.info(f"Done {num_done}, errors on {num_error}")

    def export_alignments(
        self,
        file_name: typing.Union[pathlib.Path, str],
        training_graph_archive: FstArchive,
        feature_archive: FeatureArchive,
        reference_phone_archive: RandomAccessInt32VectorVectorReader = None,
        word_file_name: typing.Union[pathlib.Path, str] = None,
        likelihood_file_name: typing.Union[pathlib.Path, str] = None,
        write_scp: bool = False,
        callback: typing.Callable = None,
        utterance_parameters: typing.Dict[str, typing.Dict[str, typing.Any]] = None,
    ) -> None:
        """
        Export alignments from training graph and features archives to an alignment archive file

        Parameters
        ----------
        file_name: str or :class:`~pathlib.Path`
            Alignment archive file path
        training_graph_archive: :class:`~kalpy.decoder.data.FstArchive`
            Archive of training graph FSTs per utterance
        feature_archive: :class:`~kalpy.feat.data.FeatureArchive`
            Archive of feature matrices per utterance
        reference_phone_archive: :class:`~_kalpy.util.RandomAccessInt32VectorVectorReader`, optional
            Archive of reference phone sequences per utterance
        word_file_name: str or :class:`~pathlib.Path`, optional
            File path to save word sequence
        likelihood_file_name: str or :class:`~pathlib.Path`, optional
            File path to save per frame log-likelihoods
        write_scp: boolean
            Flag for whether to write alignment archive with SCP file, defaults to False
        callback: callable, optional
            Callback to emit :class:`~kalpy.gmm.data.Alignment` objects as they're processed
        """
        write_specifier = generate_write_specifier(file_name, write_scp)
        writer = Int32VectorWriter(write_specifier)
        word_writer = None
        if word_file_name:
            word_write_specifier = generate_write_specifier(word_file_name, write_scp)
            word_writer = Int32VectorWriter(word_write_specifier)
        likelihood_writer = None
        if likelihood_file_name:
            likelihood_write_specifier = generate_write_specifier(likelihood_file_name, write_scp)
            likelihood_writer = BaseFloatVectorWriter(likelihood_write_specifier)
        try:
            for alignment in self.align_utterances(
                training_graph_archive,
                feature_archive,
                reference_phone_archive=reference_phone_archive,
                utterance_parameters=utterance_parameters,
            ):
                if alignment is None:
                    continue
                if isinstance(alignment, tuple):
                    if callback:
                        callback(alignment)
                    continue
                if callback:
                    callback((alignment.utterance_id, alignment.likelihood))
                writer.Write(str(alignment.utterance_id), alignment.alignment)
                if word_writer is not None:
                    word_writer.Write(str(alignment.utterance_id), alignment.words)
                if likelihood_writer is not None:
                    likelihood_writer.Write(
                        str(alignment.utterance_id), alignment.per_frame_likelihoods
                    )
        except Exception as e:
            logger.error(e)
            raise
        finally:
            writer.Close()
            if word_writer is not None:
                word_writer.Close()
            if likelihood_writer is not None:
                likelihood_writer.Close()
